/* 3 feb 2003. Daniel Hellerstein (danielh@crosslink.net)
  UpDater: An rsync based web-based file updater.
        This is the "central" component.
        It is queried by a remote client using one of several remote components.
        The client will ask for a "package" of files. 
        This central component will then send a list of these files, along with some 
        information about them (such as last modified date).  
        The client will then ask for some (or all) of these files. 
        This central component will rsync and/or gzip and/or encrypt each
        requested file, and send it back to the remote client.
        
        The "remote component" may be another SREhttp/2 server (a remote server),
        an individual using a web browser and the UpClient helper application, or
        or an individual running a standalone client (UpDateME.CMD).

       Hint: to use encryption, you'll need some extra software. In particular,
             we recommend Blowfish (check hobbes.nmsu.edu).

 For details, see updater.htm
*/

/**** User changeable parameters ****/

/* 1 to use K and M abbreviations when  displaying filesizes, 0 to show full number */
use_abbrev=1 

/* Always use cache entry if it's current. That is, suppress NOCACHE=1 option */
always_use_cache=1

/* add crc's of file in package list (required for "crc32 check" by remote
   servers. 1=yes, 0=no */
addcrc=1

/* colors and or background used in some html displays */
back1='bgcolor="#88aa66"'

/* allow use of GZIP encoding */
candogzip=1

/* allow use of Rsync-differencing encoding */
candorsync=1

/* set of "encryption" specs. ENCRYPT.!BLOWFISH is an example,
  you can delete it if you will never use blowfish
  Warning (3 feb 03): the "status checker" of RUpdater will sometimes
      cause SREhttp/2 to crash when blowfish is running. 
      You may want to try using the blowfish rexx-callable procedure instead
     (www.srehttp.org/apps/rxblfish)
  Alternatively, the SIMPLE encryption method can ALWAYS be used (it's a weak 32 bit 
  encryption).
*/
ENCRYPT.!blowfish='bf -q -e %in %out "%key"'


/* list of files, in or under the updater directory, that the SEND= option
   is allowed to "send to the client" */
send_files='UPDATER.HTM RXGZLIB.DLL GZIP.EXE RXRSYNC.DLL UPCLIENT.CMD UPDATEME.CMD'

/*  the "updater" directory -- should be a fully qualified directory.
    Or, leave it blank to use the default (the UPDATER subdirectory of 
    the addon directory of SREhttp/2 */

UPDATER_DIR=''
 
 
/**** End of User changeable parameters ****/

updater:    

/* parse arg  ddir, tempfile, reqstrg,list,verb ,uri,user, ,
          basedir ,workdir,privset,enmadd,transaction,verbose, ,
         servername,host_nickname,homedir,aparam,semqueue,prog_file,reqnum  */


parse arg list,servername,verb,tempfile,,
          prog_file,reqnum,verbose,client_ip,privset,,
          uri,host_nickname,id_info,aiter,attribs,seluse2

if verb="" then do
   say " This SREhttp/2 addon is not meant to be run in stand-alone mode. "
   exit
end  
signal on error name badguys; signal on syntax name badguys
crlf = '0d0a'x

parse var uri thisaddon '?' .

/*
if aiter='' then do
    rcode='PRIVS , 'action' addon may require privileges '
    return rcode
end 
*/

/* may need this below */
parse var privset privset1 ',' privset2
privset=privset1

/* look in privset2 for ?UPDATER= "Secret privilege" */

shared_secret='UPDATER'
do until privset2=''
   parse upper var privset2 apriv privset2
   if abbrev(apriv,'UPDATER:')=0 then iterate
   parse var apriv . ':' shared_secret
   shared_secret=strip(shared_secret)
   leave
end

if updater_dir='' then do
   basedir=sre_value('H2_ADDON_DIR_DEFAULT',,'SRE')
   updater_dir=strip(basedir,,'\')'\UPDATER'
end

if dosisdir(updater_dir)=0 then do
     foo=sre_command('header add x-updater: no updater_dir error',,id_info)
     return 'string UPDATER configuration error. No such updater directory: 'updater_dir      
end

package_dir=updater_dir'\PACKAGE'
if dosisdir(package_dir)=0 then do
     foo=sre_command('header add x-updater: no package_dir error',,id_info)
     return 'string UPDATER configuration error. No such package directory: 'package_dir
end

temp_dir=updater_dir'\TEMP'
if dosisdir(temp_dir)=0 then do
     foo=sre_command('header add x-updater: no temp dir error ',,id_info)
     return 'string UPDATER configuration error. No such temporary directory: 'temp_dir
end

call check_dlls         /* load some dlls */

/* pull out arguments */
args.=''
selects.=''
if list='' then list='package=!list&ashtml=1'
do until list=""                /* get user input */
   parse var list a1 '&' list
   parse var a1 a1a '=' a1b0
   a1b0=sre_packur(translate(a1b0,' ','+'||'0d0a00900'x))
   a1a='!'||translate(strip(a1a))
   args.a1a=strip(a1b0)
   args.!arglist=args.!arglist' 'a1a
end

call do_central

return rcode 

/*************************************/
/* here if updater is acting as a central server */
do_central:

anaction=translate(fixit(args.!ACTION))


/* bump up timeout inactive */
goo=sre_extract('LIMITTIMETOTAL',id_info)
aa=min(goo,75)-1
foo=sre_command('SET LIMITTIMEINACTIVE ',aa,id_info)

/* send UPCLIENT.CMD, etc */
if strip(translate(args.!SEND))<>'' then do
 if wordpos(strip(translate(args.!send)),translate(SEND_FILES))=0 then do
       foo=sre_command('header add x-updater: no upclient.cmd error',,id_info)
       rcode='string no such file available 'args.!send
       return 0
  end 
  afile=updater_dir'\'||strip(args.!send)
  foo=stream(afile,'c','query exists')
  if foo='' then do
     call lineout tempfile,'<html><head><title>UpClient not available</title></head>'
     call lineout tempfile,'<body 'back1'> '
     call lineout tempfile,'<h3>File not available </h3>'
     call lineout tempfile,'Sorry, 'args.!send ' is not available.'
     call lineout tempfile,'</body></html>'
     call lineout tempfile
     rcode='FILE ERASE type text/html name 'tempfile 
     return 1
   end
   else do
     select 
        when args.!text=1 then do
           rcode=sre_command('FILE  nocache  type text/plain  name 'foo,,id_info,,'Content-Disposition: attachment; filename="'args.!send'"' )
        end
        when args.!text=2 then do
          rcode=sre_command('FILE NOCACHE type text/html name 'FOO,,id_info,,'Content-Disposition: attachment; filename="'args.!send'"' ) 
        end
        otherwise do
          rcode=sre_command('FILE nocache type application/octet-stream  name 'foo,,id_info,,'Content-Disposition: attachment; filename="'args.!send'"' )
        end
     end
     return status
   end 
end 

/* edit packages ? */

if abbrev(anaction,'EDIT_PACKAGE')=1  then do
   call sec_check
   rr=result
   if rr<>1 then return rcode
end

if anaction='EDIT_PACKAGE' then do
   call edit_package
   rcode='FILE erase TYPE text/html NAME 'tempfile
   return 1
end

/* create a package ? */
if anaction='EDIT_PACKAGE_NEW' then do
   call edit_package2
   rcode='FILE erase TYPE text/html NAME 'tempfile
   return 1
end

if anaction='EDIT_PACKAGE_NEW2' then do
   call edit_package2a
   rcode='FILE erase TYPE text/html NAME 'tempfile
   return 1
end


/* modify a package ? */
if anaction='EDIT_PACKAGE_MODIFY' then do
   call edit_package3
   rcode='FILE erase TYPE text/html NAME 'tempfile
   return 1
end

/* save mods of a package ? */
if anaction='EDIT_PACKAGE_MODIFY2' then do
   call edit_package3a
   rcode='FILE erase TYPE text/html NAME 'tempfile
   return 1
end


if anaction='HELP_EDIT' then do
   call help_edit
   rcode='FILE erase TYPE text/html NAME 'tempfile
   return 1
end



/* show some documentation on upclient */
if anaction='UPCLIENTDOC' then do
       call tell_upclient
       rcode='FILE erase TYPE text/html NAME 'tempfile
       return 1
end 

/* Ask REMOTE CLIENT for some options */
if abbrev(anaction,'GETL2')=1 then do    
       call get_client_options
       rcode='FILE erase TYPE text/html NAME 'tempfile
       return rcode
end

/* verify that the client's shared secret is correct */
if strip(translate(args.!VERIFY))<>'' then do

   if privset='' then do
       rcode='PRIVS_CHECK , UpDater needs an encryption key.'
       return rcode
   end 

    parse upper var args.!verify nonce';'md4 . ; md4=strip(md4)
    tocheck=space(translate(nonce||shared_secret),0)
    md4b=translate(rx_md4(tocheck))
    if  md4=md4b then
       av='VERIFY=1'
    else
       av='VERIFY=0'
    rcode='VAR type application/x-updater '||'0d0a'x||av
    return 1
end 


/* (the default) return a list of current packages */
IF ARGS.!PACKAGE='?' | translate(strip(ARGS.!PACKAGE))='!LIST' then DO
      call do_packagelist
      ah='X-updater: package_list '
      if astext=1 then
         rcode=sre_command('FILE erase type text/plain name 'tempfile,,id_info,,ah )
      else
         rcode=sre_command('FILE erase type text/html name 'tempfile ,,id_info,,ah)
      return 1
end 

package_file=package_dir'\'||strip(args.!package)||'.PKG'
isize=stream(package_file,'c','query size')
if (isize=0 | isize='') then do
    hh='header add x-updater: error'
    rcode=sre_command('string  UpDater ERROR: empty or missing package: 'args.!package,,id_info,,hh)
    return 1
end

stat=read_package(package_file,isize)             /* read in package file to package_opts. */

/* see if any privileges required */
if (package_opts.!Privs='' | pos('*',package_opts.!Privs)>0) & package_opts.!ENCRYPT=''  then do
   nop
end 
else do               /* check privs */
   OK=check_privs(privset,package_opts.!privs,package_opts.!ENCRYPT)
   if ok<>'OK' then return rcode  /* check_privs will issue auth request */
end 


/* Otherwise, return info on a selected package */
/*  ? -- a request for listing of currently available packages
    PACKAGE_NAME -- a name of a package (must correspond to a
                    file in the PACKAGE_DIR 
*/




/* check action. */

 select

/* list contents of this package */
    when abbrev(ANACTION,'DISP') then DO                
       call disp_package
       hh='header add X-updater: display_package '
       rcode=SRE_command('FILE erase TYPE text/html NAME 'tempfile,,id_info,,hh)
       return 1
    end    

    when  abbrev(anaction,'GETL3')=1 then do    
      xlist='get=filelist3'
      call do_getl3
      ah='header add X-updater: 'xlist
      rcode=SRE_command('FILE erase TYPE text/html NAME 'tempfile,,id_info,,ah)
      return 1
   end

    when  abbrev(anaction,'GETL4')=1 then do    /* same as send_get, with extra constraints */
       xlist='get=filelist'
       if candorsync=1 then xlist=xlist||',rsync'
       if candogzip=1 then xlist=xlist||',gzip'
       selects.!ALL=0                           /* selects.!ALL='' means "no selective files" */
       if args.!ALL=1 then do
          selects.!ALL=1
       end 
       else do                  /* create a selection list  */
          do mm=1 to words(args.!ARGLIST)
               aw=strip(word(args.!arglist,mm))
               if abbrev(aw,'!F_')+abbrev(aw,'!DIR_')=0 then iterate
               if abbrev(aw,'!DIR_')=1 then do
                  if strip(args.aw)=1 then do
                     parse var aw . '_' jj .
                     selects.jj=1
                  end
               end
               if abbrev(aw,'!F_')=1 then do
                    parse var aw . '_' entry '_' . 
                    aff=args.aw ; aff2=translate('!'||aff)
                    selects.entry.aff2=1
               end 
          end          /* scan of args. */
       end             /* not ALL */

       call send_gets
       if broken=1 then return '0 '
       if astext=1 then do
           ah='header add X-updater: 'xlist
           rcode=SRE_command('FILE erase TYPE application/x-updater NAME 'tempfile,,id_info,,ah) 
       end
       else do
          ah='header add X-updater: 'xlist
          FOO=SRE_command('FILE erase TYPE text/html NAME 'tempfile,,id_info,,ah)
       end
       return 1
   end


/* return a "get list". The remote server should request each file on this
   list, subject to date checking, possibly using rsync */

    when abbrev(anaction,'GETL')=1 then do    /* set package list ! */
       xlist='get=filelist'
       if candorsync=1 then xlist=xlist||',rsync'
       if candogzip=1 then xlist=xlist||',gzip'
       call send_gets
       if broken=1 then return '0 '
       if astext=1 then do
           ah='header add X-updater: 'xlist
           FOO=SRE_command('FILE erase TYPE application/x-updater  NAME 'tempfile,,id_info,,ah) 
       end
       else do
          ah='header add X-updater: 'xlist
          FOO=SRE_command('FILE erase TYPE text/html NAME 'tempfile,,id_info,,ah)
       end
       return 1
    end

/* get a file from a package */
    when abbrev(anaction,'FILE')=1 then do   
        call get_file
        return result
    end
    otherwise 
 end

RCODE=SRE_COMMAND('string UpDater: No action was specified ',,ID_INFO,,'x-updater: error')
RETURN 1



/*************************************/
/* get client options */
get_client_options:
packagename=translate(strip(args.!package))
call lineout tempfile,'<html><head><title>Updater Client options </title></head>'
call lineout tempfile,'<body 'back1'>'
CALL LINEOUT TEMPFILE,'<h3>Please select a few UpDater  options</h3>'
call lineout tempfile,'Files listed in the <tt>'args.!package '</tt> package (on <tt> ' servername'</tt>)'
call lineout tempfile,'will be UpDated to your computer.'

   
/* create a form  */
call lineout tempfile,'<form action="'||thisaddon||'" method="GET">'
call lineout tempfile,'<input type="hidden" name="action" value="GETL">'
call lineout tempfile,'<input type="hidden" name="package" value="'args.!package'">'
call lineout tempfile,'<input type="hidden" name="client" value="1">'
if args.!GETUSER=1 then do
  call lineout tempfile,'<input type="hidden" name="GETUSER" value="1">'
end

call lineout tempfile,'<input type="Submit" value="Get the files in 'packagename'... ">'
call lineout tempfile,'<table><tr valign="top">'

if args.!GETUSER=1 then do
  call lineout tempfile,'<td valign="top">Username and password  </td>'
  call lineout tempfile,'<td>Username:<input type="text" name="USER" size=10  > '
  call lineout tempfile,'Password: <input type="password" name="PWD"  size=10> '
  call lineout tempfile,'<br><em>The username and password required for the 'packagename' package may be different'
  call lineout tempfile,' than your 'servername'  "logon" username and password</em>'
end

call lineout tempfile,'</td>'

call lineout tempfile,'<tr valign="top"><td>Destination directory (on your computer)</td> '
call lineout tempfile,'<td><input type="text" name="DESTDIR"  size=30> </td>'

call lineout tempfile,'<tr valign="top"><td>Backup directory (on your computer).</td> '
call lineout tempfile,'<td><table border=1><tr><Td><input type="text" name="BACKUPDIR" size=30>'
call lineout tempfile,'<br><em> To suppress backing up changed files, leave this field empty.</em></td></table></td>'


call lineout tempfile,'<tr valign="top"><td>Update checks &amp; cleanup option</td>' 
call lineout tempfile,'<td><input type="radio" name="CHECK" value="1" checked> '
call lineout tempfile,'Update if newer (using file creation date) <bR>'
call lineout tempfile,'<input type="radio" name="CHECK" value="2" > '
call lineout tempfile,'Update if 32bit (crc32) file hash is different <bR>'

/* call lineout tempfile,'<input type="radio" name="CHECK" value="3">'
call lineout tempfile,'Update if 128bit (md4) file hash is different <bR>' */

call lineout tempfile,'<input type="radio" name="CHECK" value="0">'
call lineout tempfile,'Always update '

call lineout tempfile,'<hr width="25%"> <input type="checkbox" name="DODELETE" value="1" > '
call lineout tempfile,'Delete all existing files (in your destination directory) that are not mentioned in this package'
call lineout tempfile,'</td>'

call lineout tempfile,'<tr><td colspan=2 align="middle"><hr width="50%"><b>Advanced Options </b><br> </td>'
call lineout tempfile,'<tr valign="top"><td valign="top">Range to retrieve</td>'
call lineout tempfile,'<td> Retrieve entries from '
call lineout tempfile,'#<input type="text" name="RANGE1"  size=5></em>'
call lineout tempfile,' to '
call lineout tempfile,'#<input type="text" name="RANGE2"  size=5><em>(leave both boxes blank to retrieve all files)</em>'
call lineout tempfile,'</td>'
call lineout tempfile,'<tr ><td valign="top">Features </td>'
call lineout tempfile,'<td><input type="checkbox" name="nocache" value=1 > Use the latest version of <tt> 'packagename'</tt> ?'
if args.!TORUN<>'' then
  call lineout tempfile,'<br><input type="checkbox" name="dorun" value=1 > Automatically execute the <tt>'args.!torun||' </tt><em>installation</em> program <em>(if available)</eM> ?'
call lineout tempfile,'<br><input type="checkbox" name="quiet" value=1 > Quiet output (minimal display of status messages)?'
call lineout tempfile,'</td>'
call lineout tempfile,''
call lineout tempfile,''

call lineout tempfile,''
call lineout tempfile,''
call lineout tempfile,'<tr><td colspan=2 align="middle"><hr width="75%"> </td>'

call lineout tempfile,'<tr><td> </td>'
ab='Get the files in 'packagename'...'
call lineout tempfile,'<td align="right"><input type="Submit" value="'||reverse(ab)||'"></td>'

call lineout tempfile,'</table><hr>'

call lineout tempfile,'</form>'

call lineout tempfile,'<p>'
call lineout tempfile,'<a href="'||thisaddon||'?">Return to UpDater main menu</a> (at 'servername')'

call lineout tempfile,'</body></html>'

call lineout tempfile


return 1


/*************************************/
/* get a file from a package
   Uses filename, entry #, and package name to determine 
   actual file.
   Assuming that no_encode does not apply to this file,
      If RSYNC specified, will return an rsync differenc file.
      If Gzip specified, will gzip the file
     (Gzip after RSYNC).
      If encryption specified, encrypt it (after gzip)
      
*/   
get_file:

afile=strip(translate(args.!file,'\','/'),,'\')

aentry=strip(args.!entry)
doerase=' ' ;celist=''
/* say " package_opts.0 " package_opts.0  */
if aentry> package_opts.0 then do
        string "No such entry: "aentry
        exit ' '
end 
if Pos('\',afile)>0 then do
    if package_opts.aentry.!subs<>1 then do
        string "Subdirectories are not accessible "
        exit ' '
    end 
end 

/* append afile to !cfile, but watch out for wildcards */
cfile=package_opts.aentry.!cfile

foo=translate(cfile,' ','\')
aa=word(foo,words(foo))
if (pos('.',aa)+pos('*',aa)+pos('?',aa))=0 then cfile=cfile'\*.*'

iast=pos('*',cfile)
if iast>0  then do
   islash=lastpos('\',cfile,iast)
   cfile2=left(cfile,islash)||afile
end
else do
   cfile2=cfile
end


/* excluded file? */
aa=check_exclude(cfile2)
if aa=1 then do
   rcode=sre_error_response('forbid',,,'Updater can not access 'afile' in package: 'args.!package)  
   return 1
end 
aa=sysfiletree(cfile2,'goo','FT')

if goo.0=0 then do              /* couldn't find it */
   rcode=sre_error_response('notfound',,,'Updater could not find 'afile' in package: 'args.!package)  
   return 1
end

if anaction='FILE1' then do    /* force without encoding or date checking */
  mtype=sreh2_get_mimetype(cfile2,host_nickname)
  ah='header add content-disposition:  attachment; filename="'||filespec('n',cfile2)||'"'
  rcode=sre_command('file type 'mtype' name 'cfile2,,id_info,,ah)
  return foo
end

parse var goo.1 adate orig_size .

xlist='SIZE='||strip(orig_size)||',DIR='||strip(package_opts.aentry.!rdir)

/* note: the date check and the hash check are typically NOT done
  by updater (they are done by rupdater or upclient, using info
  in the package-listing file */

/* check modification date -- no longer used, instead check on client side */
if (args.!mod<>''  ) then do 
  if left(adate,2)<70 then 
    adate='20'||adate
  else
    adate='19'||adate
  if (adate <= args.!mod) then do       /* unmodified */
        ah='NoAuto Add Date: '||sre_gmt()
        dasoft=sre_server('NF')
        ah=ah||'0d0a'x||'Server: '||dasoft
        ah=ah||'0d0a'x||'X-UpDater: 'xlist
        rcode=sre_command('NODATA  NOWAIT ',,id_info,'304 Not Modified',ah)
        return 1
  end
end

/*  -----------------
 check hash? -- no longer used (instead, crc32 check on client side 
if args.!md4<>''  then do
   amd4=strip(translate(rx_md4(cfile2,1)))
   if amd4=strip(translate(args.!Md4)) then do
        ah='NoAuto Add Date: '||sre_gmt()
        dasoft=sre_server('NF')
        ah=ah||'0d0a'x||'Server: '||dasoft
        ah=ah||'0d0a'x||'X-UpDater: 'xlist
        rcode=sre_command('NODATA  NOWAIT ',,id_info,'304 Not Modified',ah)
        return 1
   end
end
 --------------------- */


/* are we allowed to encode this file? */
if strip(args.!No_encoding)=1 then
 noencode=1
else
  noencode=check_noencode(cfile2)


/* rsync this? */
if  noencode=0 & candorsync=1 then do 
  call rsignature
  rr=result
  if rr=1 then do   /* found signature */
       dorsync=do_rsync(rheader,cfile2)
       if dorsync<>0 then do
           cfile2=dorsync
           xlist=xlist||',md4='||amd4
           doerase='ERASE '     
           celist='gdiff'
           foo=stream(cfile2,'c','close')
       end 
  end
end

lencfil2=stream(cfile2,'c','query size')
/* gzip this ? */
if args.!gzip=1 & candogzip=1 & noencode=0 & lencfil2>100 then do
  stat=do_gzip(cfile2)
  if stat<>0 then do
      if doerase='ERASE' then do
        foo=sysfiledelete(cfile2)
      end
      doerase='ERASE'
      cfile2=stat
      celist=strip(celist||',gzip',,',')
   end
end
/* encrypt? */
if args.!ENCRYPT<>'' & package_opts.!encrypt<>'' then do
  useenc=''
  aencrypts=translate(strip(translate(args.!encrypt,' ',',')))
  do ll=1 to words(aencrypts)  /* use 1st that matches package_opts.!encrypts */
      a1=strip(word(aencrypts,ll))
      if wordpos(a1,package_opts.!encrypt)>0 then do
         useenc=a1
         leave
      end 
  end
  if useenc='' | useenc='#SIMPLE' then do
        useenc='#SIMPLE'  /*  no agreeable encryption, use Simple method */
        encline='#SIMPLE'
  end
  else do
    aa='!'useenc
    encline=encrypt.aa
  end
  cfile2=do_encrypt(cfile2,encline,doerase)
  if cfile2='' then do               /* error, give up */
        if doerase='ERASE' then foo=sysfiledelete(cfile2)
        ah='NoAuto Add Date: '||sre_gmt()
        dasoft=sre_server('NF')
        ah=ah||'0d0a'x||'Server: '||dasoft
        ah=ah||'0d0a'x||'X-UpDater: Encryption failure using '||useenc
        rcode=sre_command('NODATA  NOWAIT ',,id_info,'500 Internal Server Error',ah)
        return 1
  end 
  if celist='' then
     celist='ENCRYPT='useenc
  else
     celist=celist',ENCRYPT='useenc
  doerase='ERASE'
end 

ah='x-updater: 'xlist

if celist<>'' then do
  ah=ah||'0d0a'x||'content-encoding: 'celist
end

rcode=sre_command('file 'doerase ' type application/x-updater name 'cfile2,,id_info,,ah)

return 1

/***********************************/
/* encrypt a file */
do_encrypt:procedure expose  temp_dir shared_secret thisaddon
parse arg oldfile,doit,doerase

atid=dostid()
apid=dospid()

tmpf=temp_dir'\G'||strip(left(apid,3))||atid||'.ENC'
foo=sysfiledelete(tmpf)
if doit='#SIMPLE' then do               /* use simple method */
   ff=do_simple_enc(oldfile,tmpf,shared_secret)
end 
else do
  doit=sre_replacestrg(doit,'%IN',oldfile,'ALL')
  doit=sre_replacestrg(doit,'%OUT',tmpf,'ALL')
  doit=sre_replacestrg(doit,'%KEY',shared_secret,'ALL')
  doit=strip(doit)
  aa=word(doit,1)

  signal on syntax name noenc ; signal on  error name noenc

  if pos('(',aa)>0 then do
    doit='foo='doit
    interpret doit 
  end
  else do
     doit='@'||doit|| ' 2>nul '
     address cmd doit
  end 

  signal on error name badguys; signal on syntax name badguys
end

if doerase='ERASE' then foo=sysfiledelete(oldfile)
doerase='ERASE'

return tmpf             /* the file name to use */

noenc:
signal on error name badguys; signal on syntax name badguys
foo=sysfiledelete(tmpf)

noenc:
return ''             /* failure, send error message (401) */


/* ======== */
do_simple_enc:procedure
parse arg afile,outfile,akey

instuff=stream(afile,'c','query size')
stuff2=charin(afile,1,instuff)
foo=stream(afile,'c','close')

aa=stringcrc(akey)

cc=c2x(aa)
ix=max(2,x2d(substr(cc,1,2)))
iy=max(2,x2d(substr(cc,3,1)))
iz=max(2,x2d(substr(cc,4,1)))

numeric digits 12
j4=4
mx32=4294967295

amask=''
irepeat=20+(iy//28)
ii=0
do forever
  arand=random3(mx32)
  darand=right(d2c(arand),j4,0)
  amask=darand||amask||darand
  if (ii//irepeat)=1 then amask=amask||amask
  if length(amask)>instuff then leave
  ii=ii+1
end

aanew=bitxor(stuff2,left(amask,instuff))               /* THIS IS THE ENCRYPTED RESULT !!!  */

/* aanew2=bitxor(aanew,left(amask,instuff))  */              /* THIS IS THE DECRYPTED RESULT !!!  */

foo=charout(outfile,aanew,1)
foo=stream(outfile,'c','close')
return 1


/* *********************** */
/* UNIFORM DISTRIBUTION RANDOM # GENERATOR.
  FROM APPL STATIS 1982, VOL31 Pg183 */

random3:procedure expose ix iy iz
parse arg mx32
IX=(171*IX)//30269
IY=(172*IY)//30307
IZ=(170*IZ)//30323
RANDOM=(IX/30269.) + (IY/30307.)  + (IZ/30323)
random=trunc((random // 1.0)*mx32)

RETURN random




/***********************************/
/* gzip a file */
do_gzip:procedure expose temp_dir  thisaddon
parse arg afile

atid=dostid()
apid=dospid()

tmpf=temp_dir'\G'||strip(left(apid,3))||atid||'.GZ'
rc=rxgzDeflateFile(afile,tmpf)
if rc=0 then return tmpf
foo=sysfiledelete(tmpf)
return 0

/***************/
/* compute a gdiff difference file, using rsync-signature as a synopsis */
do_rsync:Procedure expose amd4 temp_dir  thisaddon
parse arg rsign,ffile

rsign=sre_pack64(rsign)
blocksize=c2d(substr(rsign,1,4))
nblocks=c2d(substr(rsign,5,1))
synopsis=''
do mm=1 to nblocks
   jat=((mm-1)*8)+6
   foo=substr(rsign,jat,8)
   synopsis=synopsis||left(foo,20)
end

comment='srersync pre-reply '
aa=left(comment,81)||left(blocksize,7)||left(nblocks,9)||left('md4',35)
synopsis=aa||synopsis

atid=dostid()
apid=dospid()
a1=temp_dir||'\'atid'_'apid||'.SYN'
foo=sysfiledelete(a1)
status=rx_rsync_gdiff(ffile,synopsis,a1,1)     /* compute a gdiff, using sypnopsis */
foo=stream(a1,'c','close')
parse var status status amd4 ; status=strip(status)
if status<>'OK' then do
  if verbose>1 then call pmprintf('UpDater rsync error: 'status)
  foo=sysfiledelete(a1)
  return 0              /* error, could not do rsync */
end

foo=stream(a1,'c','open read')
if abbrev(strip(translate(foo)),'READY')=0 then do
  foo=stream(a1,'c','close')
  foo=sysfiledelete(a1)
  if verbose>1 then  call pmprintf('UpDater WARNING: unable to open difference file 'a1)
  return 0
end

issize=stream(a1,'c','query size')
if (issize=0 | issize='') then do       /* some kind of error */
  foo=stream(a1,'c','close')
  foo=sysfiledelete(a1)
  return 0
end

return a1               /* return file containing rsync result */


/****************/
/*Look for a rsync-signature request header. If none, do nothing */
rsignature:
rheader=sre_reqfield('rsync-signature_updater',,id_info)
if rheader='' then return 0
return 1


/*************************************/
/* create a "get list". This list will be used by remote server, 
or remote client, or a remote standalone client.
The remote app should ask for each of these files; possibly checking
datestamps, etc first, and perhaps using rsync & gzip  */

Also, possibly use cache file, or save as cache file.

See updater.htm for details on the structure of these "list of files"

*/
send_gets:                              /* return # of entries written */

/* find all files matching each  .!cfile entry */

/* either send as plain text (for remote server processing), or
  as an html list (for browser-based human processing */
astext=1
iserror=0
broken=0
do_multi=0


/* if remote client (via updater), check for required fields first */
if args.!client=1 & args.!fst<>1 then do
  ddir=fixit(args.!destdir)
  if ddir='' then do
    call lineout tempfile,'<html><head><title>ERROR: no destination directory </title></head>'
    call lineout tempfile,'<body 'back1'> '
    call lineout tempfile,'<h3>ERROR: no destination directory </h3>'
    call lineout tempfile,'You forgot to specify a destination directory.'
    call lineout tempfile,'The destination directory is a directory on <b>your</b> computer,'
    call lineout tempfile,'it is where UpDater will place the files in this package.'
    call lineout tempfile,'<p><em>Please use your browser''s BACK key, and select a destination directory</em>'
    call lineout tempfile,'</body></html>'
    call lineout tempfile
    astext=0
    return 1
  end
  if args.!GETUSER=1  then do
      auser=fixit(args.!user) 
      apwd=fixit(args.!pwd) 

  if (package_opts.!privs='' | pos('*',package_opts.!privs)>0)  & package_opts.!ENCRYPT='' then

      if (auser='' | apwd='' ) & (package_opts.!privs<>'' ) & (pos('*',package_opts.!privs)>0)  then do
        call lineout tempfile,'<html><head><title>ERROR: no username/password </title></head>'
        call lineout tempfile,'<body 'back1'> '
        call lineout tempfile,'<h3>ERROR: no username or no password </h3>'
        call lineout tempfile,'You forgot to specify a username and password.'
        call lineout tempfile,'To UpDate files in this package, you must have an appropriate <b>username and password</b>.'
        call lineout tempfile,'This <em>package specific</em> username and password may, or may <b>not</b>, be the same as the <em>logon username and password</em> required by this server.'
        call lineout tempfile,'<p><em>Please use your browser''s BACK key, and select a username and password</em>'
        call lineout tempfile,'</body></html>'
        call lineout tempfile
        astext=0
        return 1
      end 
  end 
end 

numeric digits 11
nowtime=date('b')+(time('m')/1440)    /* current date/time  -- 1 minute resolution */
parse var thisaddon areq '?' .

/* possibly return a cached filelist */
do yi=1 to 1
   if selects.!ALL<>'' then leave       /* if custom selection, then can't use cache */
   if strip(args.!ashtml)=1 | package_opts.!cache=0 then leave
   if (always_use_cache<>1 & args.!nocache=1) | package_opts.!cachestamp=0 then leave

   if package_opts.!cachestamp+package_opts.!cache<nowtime then leave  

/* IF here, use cached entry! */
   goo=stream(package_opts.!cachefile,'c','open read')
   if abbrev(strip(translate(goo)),'READY')=0 then leave
   sz3=stream(package_opts.!cachefile,'c','query size')
   if sz3=0 | sz3='' then do
      if verbose>1 then call pmprintf(' UpDater warning: empty cache file: 'package_opts.!cachefile)
      leave
   end
   aax=charin(package_opts.!cachefile,1,sz3)
   foo=stream(package_opts.!cachefile,'c','close')
   parse var aax hdrline '0d0a'x aax
   if length(aax)=0 then leave
   aline='Package='args.!package
   aline=aline||' '||'Server='servername
   if package_opts.!encrypt<>"" then aline=aline' Encrypt='||translate(package_opts.!encrypt,',',' ')
   
   CC=GET_CREATED(HDRLINE)
   parse var cc cc1 cc2

   tnow=make_timestamp(package_file)

   if tnow>cc2 then leave         /* if package file has changed, don't use cache */
   aline=aline||' '||'CREATED='CC1
   aline=aline||' '||'pKG_DATE='CC2
   if package_opts.!Privs<>'' & pos('*',package_opts.!Privs)=0 then 
          aline=aline||' NEEDPRIVS=1'
   if candogzip=1 then aline=aline' GZIP=1'
   if candorsync=1 then aline=aline' RSYNC=1'
   aline=aline' CURRENT_TIME='nowtime
   aline=aline' SELECTOR='areq

   if args.!client=1 then call add_remote_client_stuff

   call charout tempfile,aline||'0d0a'x
   call charout tempfile,aax
   if verbose>2 then say "UpDater: using cache " package_opts.!cachefile
   foo=stream(tempfile,'c','close')
   return 1
end

cachefile=''            /* '' means "do not save to cache file */

/* NOT from cache... */
if strip(args.!ashtml)=1 then do                /* UPDATER interactive mode? */
  astext=0
  call lineout tempfile,'<html><head><title>Select a file to retrieve </title></head>'
  call lineout tempfile,'<body 'back1'> '
  call lineout tempfile,'<h3>UpDater: View &amp; retrive files (one at a time) </h3>'
  CALL LINEOUT TEMPFILE,'Select a file to retrieve from the <b>'||strip(translate(args.!package))'</b> package'
  call lineout tempfile,'<em>(on 'servername') </em> '
  call lineout tempfile,'<hr width="30%">'
  call lineout tempfile,'<em>There are <a name="top"> <tt>'package_opts.0'</tt> </a>sections in 'args.!package':</em>'
  call lineout tempfile,'<table><tr>'
  mn=0
  do jj=1 to package_opts.0
     mn=max(mn,length(package_opts.jj.!rdir))
  end /* do */
  mn=mn+3
  ncols=trunc(80/mn)
  do jj=1 to package_opts.0
       call lineout tempfile,'<td><a href="#sec'||jj||'">'package_opts.jj.!rdir'</a>&nbsp;&nbsp;&nbsp;</td>'
       if jj//ncols=0 then call lineout tempfile,'<tr>'
  end /* do */
  call lineout tempfile,'</table><hr width="30%">'
  xtras=''
  if args.!rsync=1 then xtras='&rsync=1'
  if args.!gzip=1 then xtras=xtras||'&gzip=1'
end
else do                         /* HEADER LINE FOR "list of files" */
   aline='Package='args.!package
   aline=aline||' '||'Server='servername
   if package_opts.!encrypt<>"" then aline=aline' Encrypt='||translate(package_opts.!encrypt,',',' ')
   aline=aline||' '||'CREATED='nowtime
   if candogzip=1 then aline=aline' GZIP=1'
   if candorsync=1 then aline=aline' RSYNC=1'
   if package_opts.!Privs<>'' & pos('*',package_opts.!Privs)=0 then 
          aline=aline||' NEEDPRIVS=1'
   aline=aline' SELECTOR='areq
   aline=aline||' '||'PKG_DATE='package_opts.!PKG_DATE
   if args.!client=1 then call add_remote_client_stuff

   call lineout tempfile,aline

/* save to "cached" of cachefile */  
   if astext=1 & package_opts.!cache<>0 & selects.!ALL='' then do
         foo=sysfiledelete(package_opts.!cachefile)
         foo=stream(package_opts.!cachefile,'c','open write')
         if abbrev(strip(translate(foo)),'READY')=1 then do  /* cache file open */
            cachefile=package_opts.!cachefile
            call lineout cachefile,aline 
         end 
    end


end /* do */

/* now go to each entry in package, find files, toss exclusions, and write info to
"list of files" */

ndid=0
do jj=1 to package_opts.0
   rdir=strip(package_opts.jj.!rdir)
   if (astext=0) then do
      if jj>0 then
         call lineout tempfile,'<font size="+2"><a name="sec'||jj'">'rdir'</a></font> (<a href="#top">top</a>)'  
      else
         call lineout tempfile,'<font size="+2"><a name="sec'||jj'">'rdir'</a></font> '  
      call lineout tempfile,'<ul>'  
   end
   sss=package_opts.jj.!subs
   dirget=package_opts.jj.!cfile

   foo=translate(dirget,' ','\')
   aa=word(foo,words(foo))
   if (pos('.',aa)+pos('*',aa)+pos('?',aa))=0 then dirget=dirget'\*.*'

   thisdir=strip(filespec('d',dirget)||filespec('p',dirget),,'\')

/* make sure directory exists */
   if dosisdir(thisdir)=0 then do
       call pmprintf('UpDater warning: 'args.!package', no such directory 'package_opts.jj.!cfile )
   end /* do */

   nthisdir=length(thisdir)
   csss=''; if sss=1 then csss='S'
   a1=sysfiletree(dirget,flist,'FT'||CSSS)
   ndid=ndid+flist.0
   do jj2=1 to flist.0
        parse var flist.jj2 adate asize . aname ;aname=strip(aname)
        if left(adate,2)<70 then 
            adate='20'||adate
        else
            adate='19'||adate

        KK=CHECK_EXCLUDE(ANAME)
        IF KK=1 then ITERATE

        aname2=delstr(aname,1,nthisdir)
        aname3=translate(aname2,'/','\')

/* if "client selection of files" (getl4), see if this file is permitted */
        if selects.!ALL=0 then do               /* not all selected */
           if selects.jj<>1 then do        /* not all in dir selected */
                goo2=translate('!'aname3)
                if selects.jj.goo2<>1 then iterate /* this file is NOT selected, so iterate */
           end 
        end 

/* HERE IS WHERE A CLIENT (non-html) LIST IS CONSTRUCTED */
        if astext=1 then do
           acrc=''
           noce=check_noencode(aname)
           if addcrc=1 then do
               acrc=fileadler(aname)
           end
           lline=adate||' '||jj||' '||aname3||' 'rdir' 'noce' 'acrc' 'asize

           if do_multi=1 then do
              rcode=sref_multi_send(lline||'0d0a'x,,'1A',0)
              if rcode<0 then do
                if verbose>1 then call pmprintf('UpDater: connection broken ('rcode')?')
                 broken=1 
                 return 1
              end 
           end 
           else do
              call lineout tempfile,lline
           end
           if cachefile<>'' then do
              call lineout cachefile,adate||' '||jj||' '||aname3||' 'rdir' 'noce' 'acrc' 'asize
           end

        end

/* html response */
        else do         
           modddate=''
           if args.!nodate<>1 then moddate='&mod='||adate
           aurl='<a target="UpDater" href="'||thisaddon||'?file='||aname3||'&entry='||jj||'&package='||args.!package||'&action=file1' ,
                ||xtras||'">'
           call lineout tempfile,'<li> '||aurl||aname3||'</a> '
        end /* do */

   end /* do */
   if (astext=0) then do
      call lineout tempfile,'</ul>'  
   end
end /* do */
if astext=0 then do
        call lineout tempfile,'<a href="'||thisaddon||'?">Return to updater main menu</a> (at 'servername')'
        call lineout tempfile,'</body></html>'
end /* do */
call lineout tempfile

/* close cache file? */
if cachefile<>'' then foo=stream(cachefile,'c','close')

return ndid




/*************************************/
/*  create a selection list */

do_getl3:                             

/* find all files matching each  .!cfile entry */

/* send  as an html list (for browser-based human processing */
iserror=0
broken=0
do_multi=0


call lineout tempfile,'<html><head><title>Select  files to retrieve </title></head>'
call lineout tempfile,'<body 'back1'> '
  call lineout tempfile,'<h3>UpDater: Selective file &amp; directory retrieval  </h3>'
  CALL LINEOUT TEMPFILE,'<em>For the <b>'||strip(translate(args.!package))'</b> package'
  call lineout tempfile,'(on 'servername') </em> '

  call lineout tempfile,'<FORM action="'||thisaddon||'" method="POST" name="DO_GETL3">'
  call lineout tempfile,'<INPUT TYPE="hidden" name="ACTION" value="GETL4">'
  call lineout tempfile,'<input type="hidden" name="PACKAGE" value="'||args.!package'">'
  call lineout tempfile,'<input type="hidden" name="CLIENT" value="1">'
  if args.!GETUSER=1 then do
    call lineout tempfile,'<input type="hidden" name="GETUSER" value="1">'
  end

  if candorsync=1 then 
     call lineout tempfile,'<input type="hidden" name="RSYNC" value="1">'

  if candogzip=1 then
     call lineout tempfile,'<input type="hidden" name="GZIP" value="1">'
 
/* start of visible components of the form */
  
  call lineout tempfile,'<hr width="30%">'

  call lineout tempfile,'<table><tr valign="top">'
 
  call lineout tempfile,'<td colspan=2 align="MIDDLE"><a name="top"><input type="Submit" value="Get the selected files from 'args.!package'... "></a><br>'
  packagename=args.!package
if args.!GETUSER=1 then do
  call lineout tempfile,'<tr><td valign="top">Username and password  </td>'
  call lineout tempfile,'<td>Username:<input type="text" name="USER" size=10  > '
  call lineout tempfile,'Password: <input type="password" name="PWD"  size=10> '
  call lineout tempfile,'<br><em>The username and password required for the 'packagename' package may be different'
  call lineout tempfile,' than your 'servername'  "logon" username and password</em>'
end


  call lineout tempfile,'<tr valign="top"><td>Destination directory (on your computer)</td> '
  call lineout tempfile,'<td><input type="text" name="DESTDIR"  size=30> </td>'

  call lineout tempfile,'<tr valign="top"><td>Backup directory (on your computer).</td> '
  call lineout tempfile,'<td><table border=1><tr><Td><input type="text" name="BACKUPDIR" size=30>'
  call lineout tempfile,'<br><em> To suppress backing up changed files, leave this field empty.</em></td></table></td>'
  call lineout tempfile,' <tr valign="top"><td colspan=2 bgcolor="#aaaa33">&nbsp;</td>' 

  call lineout tempfile,' <tr valign="top"><td bgcolor="lightgrey">' ,
     '<input type="checkbox" name="ALL" value="1"> <em>Retrieve all the files</em></td>'

  call lineout tempfile,'<td>'
  call lineout tempfile,'If you retrieve all the files:'
  call lineout tempfile,'the <em>file checkboxes</em> and the'
  call lineout tempfile,'<em>section checkboxes</em> will be ignored.<br>'
  call lineout tempfile,'If you check a <b>section</b>  '
  call lineout tempfile,'the <em>file checkboxes</em> under that <b>section</b> will be ignored.'


  call lineout tempfile,' </td><tr>'

  call lineout tempfile,'<td><em>Or ... retrieve all the files from several of the <tt>'package_opts.0'</tt> sections in 'args.!package'</em>'
  call lineout tempfile,'</td>'

  call lineout tempfile,'<td bgcolor="lightgrey">'
  call lineout tempfile,'<table border=1 ><tr>'
  mn=0
  do jj=1 to package_opts.0
     mn=max(mn,length(package_opts.jj.!rdir))
  end
  mn=mn+3
  ncols=trunc(60/mn)

  ckl=''; multis=''
  do jj=1 to package_opts.0
       nowrdir=package_opts.jj.!rdir
       if wordpos(translate(nowrdir),ckl)>0 then 
          multis=multis' 'nowrdir
       else
         ckl=ckl' 'nowrdir
  end
  do jj=1 to package_opts.0
       call lineout tempfile,'<td>' 
       call lineout tempfile,'<input type="checkbox" name="DIR_'jj'" value="1">'
       nowrdir=package_opts.jj.!rdir
       if wordpos(translate(nowrdir),multis)>0 then do
          call lineout tempfile, '<a href="#sec'||jj||'">'nowrdir'(<font size=-1>'jj'</font>)</a>&nbsp;&nbsp;&nbsp;</td>'
       end
       else do
          call lineout tempfile, '<a href="#sec'||jj||'">'nowrdir'</a>&nbsp;&nbsp;&nbsp;</td>'
       end
       if jj//ncols=0 then call lineout tempfile,'<tr>'
  end 
  

  call lineout tempfile,'</table>'
  call lineout tempfile,'</td>'
  call lineout tempfile,' <tr valign="top"><td colspan=2 bgcolor="#aaaa33">&nbsp;</td>' 

  call lineout tempfile,'</table><br>'

  call lineout tempfile,'<em>Or ... select specific files to update .... </em><br>'

/* now start displaying list of directories */
ndid=0
do jj=1 to package_opts.0


   rdir=strip(package_opts.jj.!rdir)

/* write subdir name */
/*   call lineout tempfile,'<input type="checkbox" name="DIR_'jj'" value="1">'*/
   if wordpos(translate(rdir),multis)>0 then do
      call lineout tempfile,'<font size="+2"><a name="sec'||jj'">'rdir'(<font size=-1>'jj'</font>)</a></font> (<a href="#top"><font size=-1>top</font></a>)'  
   end
   else do
      call lineout tempfile,'<font size="+2"><a name="sec'||jj'">'rdir'</a></font> (<a href="#top"><font size=-1>top</font></a>)'  
   end

/* prepare to writes files in this subdir */
   call lineout tempfile,'<ul>'  
   sss=package_opts.jj.!subs
   dirget=package_opts.jj.!cfile

   foo=translate(dirget,' ','\')
   aa=word(foo,words(foo))
   if (pos('.',aa)+pos('*',aa)+pos('?',aa))=0 then dirget=dirget'\*.*'

   thisdir=strip(filespec('d',dirget)||filespec('p',dirget),,'\')

/* make sure directory exists */
   if dosisdir(thisdir)=0 then do
       call pmprintf('UpDater warning: 'args.!package', no such directory 'package_opts.jj.!cfile )
   end /* do */

   nthisdir=length(thisdir)
   csss=''; if sss=1 then csss='S'
   a1=sysfiletree(dirget,flist,'FT'||CSSS)
   ndid=ndid+flist.0

   do jj2=1 to flist.0          /* start writing files in this subdir */
        parse var flist.jj2 adate  asize . aname ;aname=strip(aname)
        if left(adate,2)<70 then 
            adate='20'||adate
        else
            adate='19'||adate

        csize=fixup_size(asize,use_abbrev)

        parse var adate aday +10 . +1 atime
        KK=CHECK_EXCLUDE(ANAME)
        IF KK=1 then ITERATE
        aname2=delstr(aname,1,nthisdir)
        aname3=translate(aname2,'/','\')
        
        modddate=''
        aurl='<input type="checkbox" name="F_'jj'_'jj2'" value="'||aname3||'">'
        call lineout tempfile,'<li> '||aurl||'  '||aname3||'&nbsp;&nbsp;<tt>'csize'  'aday' '||translate(atime,':','/')||'</tt>'

   end /* do */
   call lineout tempfile,'</ul>'  
end /* do */
  call lineout tempfile,'<br><input type="Submit" value="Get the selected files from 'args.!package'... ">'
call lineout tempfile,'<input type="reset" value="Reset the form">'
call lineout tempfile,'</form>'

call lineout tempfile,'<a href="'||thisaddon||'?">Return to updater main menu</a> (at 'servername')'
call lineout tempfile,'</body></html>'
call lineout tempfile

return ndid



/********/
/* pull stuff from cache header line, save in hdr.! */
get_CREATED:procedure expose hdr.  thisaddon
parse arg aline0
created=0 ; pkg_date=0
DO UNTIL ALINE0=''
  parse var aline0 a1 aline0
  parse var a1 avar '=' aval; avar=strip(translate(avar))
  if avar='CREATED' then created=strip(aval)
  if avar='PKG_DATE' then pkg_date=strip(aval)
end
return created' 'pkg_date
return 0

/************/
/* add remote client stuff to "aline" */
add_remote_client_stuff:
      if auser<>'' then do
         aa=auser':'apwd
         aline=aline' USERPWD='aa
      end
      if ddir<>'' then do
         aline=aline' DESTDIR='||ddir
      end
      if fixit(args.!BACKUPDIR)<>'' then do
         aline=aline' BACKUPDIR='||fixit(args.!BACKUPDIR)
      end
      if args.!DODELETE<>'' then do
         aline=aline' DODELETE='args.!DODELETE
      end
      if args.!CHECK<>'' then do
         aline=aline' CHECK='args.!CHECK
      end
      rng1=strip(fixit(args.!range1))
      if rng1<>'' then do
         rng2=strip(fixit(args.!range2))
         if datatype(rng1)='NUM' & datatype(rng2)='NUM' then do
             rng12=strip(rng1)||':'||strip(rng2)
             aline=aline' RANGE='||rng12
         end /* do */
      end
      if args.!QUIET=1 then aline=aline' QUIET=1 '
      if args.!dORUN=1 then  do
           if package_opts.!RUN<>'' & package_opts.!RUN<>'0' then do
               parse var package_opts.!run aprog atype runargs
               ctype=translate(strip(atype))
               istype=''
               if ctype='PM' then do
                  istype='PM'
               end /* do */
               else do
                  parse var package_opts.!run . runargs
               end /* do */
               isargs=''
               if abbrev(translate(strip(runargs)),'ARG=')=1 then do
                  parse var runargs . '=' isargs
               end /* do */
               aline=aline' RUN='||strip(aprog)
               if istype<>'' then aline=aline' RUNTYPE='||istype
               if isargs<>'' then aline=aline' RUNARGS='||sre_pack64_make(isargs)
           end /* do */
       end

      return 1
   

/************/
/* clean up an argument */
fixit:procedure
parse arg aa
aa=translate(aa,' ','+'||'0d0a0900'x)
aa=sre_packur(aa)
aa=strip(aa)
return aa


/*************/
/* CHECK A FILEname against an exclusion list */
check_exclude:procedure expose package_opts.  thisaddon
parse arg afile
foo=0
do mm=1 to package_opts.!exclude.0
   ex1=package_opts.!exclude.mm
   sub1=package_opts.!exclude.mm.1
 
   if (sub1<>1) then do         
      adir=translate(filespec('d',afile)||filespec('p',afile))
      xdir=translate(filespec('d',ex1)||filespec('p',ex1))
      if adir<>xdir then iterate   /* skip if not same directory */
      aname=filespec('n',afile)
      xname=filespec('n',ex1)
      foo=sre_wild_match(aname,xname)
    end
    else do             /* subdir matches okay */
      foo=sre_wild_match(afile,ex1)
    end                 /* sub1 */
    if foo<>0 then return 1           /* match found, so exclude */
end /* do */
return 0                /* if here, no exclusions */

/*************/
/* CHECK A FILEname against an  noencode list */
check_noencode:procedure expose package_opts.  thisaddon
parse arg afile

foo=0
do mm=1 to package_opts.!noencode.0
   ex1=package_opts.!noencode.mm
   sub1=package_opts.!noencode.mm.1
 
   if (sub1<>1) then do         
      adir=translate(filespec('d',afile)||filespec('p',afile))
      xdir=translate(filespec('d',ex1)||filespec('p',ex1))
      if adir<>xdir then iterate   /* skip if not same directory */
      aname=filespec('n',afile)
      xname=filespec('n',ex1)
      foo=sre_wild_match(aname,xname)
    end
    else do             /* subdir matches okay */
      foo=sre_wild_match(afile,ex1)
    end                 /* sub1 */
    if foo<>0 then return 1           /* match found, so do not encode */
end /* do */
return 0                /* if here, no noencodes */



/*************/
/* construct a list of packages.
   First line: # of packages
   2.,#Packages+1 lines: package name descripton (if requested)
*/


do_packagelist:

astext=1
if strip(args.!ashtml)=1 then astext=0
if astext=0 then do
  call lineout tempfile,'<html><head><title>Currently available UpDater packages </title></head>'
  call lineout tempfile,'<body 'back1'>'
  call lineout tempfile,'<font size=-1><a href="'||thisaddon||'?action=EDIT_PACKAGE">Edit packages</a></font><br>'
  call lineout tempfile,'<h3>Welcome to UpDater (ver 1.20)</h3>'
end

 foo=sysfiletree(package_dir'\*.pkg',plist,'FT')
 if astext=1 then do
   call lineout tempfile,plist.0
 end
 else do
    call lineout tempfile,'<blockquote>This is the UpDater <em>central server </em> &nbsp; (at <tt> 'servername')</tt> '
    call lineout tempfile,'<p>Please select one of these <tt>'plist.0'</tt> currently supported  <em>packages</em></blockquote>'
    call lineout tempfile,'<table>'
 end

 do mm=1 to plist.0
       adesc=''
       parse var plist.mm adate . . nname ;nname=strip(nname)
       if left(adate,2)<70 then 
            adate='20'||adate
       else
            adate='19'||adate
       nname0=filespec('n',nname)
       if strip(translate(args.!package))='!LIST' then do  /* extract description */
            isize=stream(nname,'c','query size')
            if (isize=0 | isize='') then do    /* bad package */
                if astext=1 then
                   call lineout tempfile,'; unavailable package: 'nname0
                else
                   call lineout tempfile,'<tr><td>'mm'</td><td><tt>'nname0'</tt></td><td><em> unavailable </em> </td>'

                iterate
            end
            stat=read_package(nname,isize)             /* read in package file */
            foo=stream(nname,'c','close')
            adesc=package_opts.!title
       end 
       nname0a=nname0
       foo=lastpos('.',nname0)
       if foo>1 then nname0a=left(nname0,foo-1)
       if astext=1 then do
          call lineout tempfile,adate' 'nname0a' 'adesc
       end
       else do
          call lineout tempfile,'<tr><td>'mm') </td><td><a href="'||thisaddon||'?package='nname0a'&action=display">'nname0a'</a> </td><td> 'adesc '</td>'
       end
 end /* do */
 if astext=0 then do
    call lineout tempfile,'</table>'
    call lineout tempfile,'<hr>Would you like to read the <a target="info" href="'||thisaddon||'?send=UPDATER.HTM&text=2">UpDater documentation</a>'
    call lineout tempfile,'</body></html>'
 end
 call lineout tempfile
 return 1

/*************************************/
/* choose a package to edit, or create a new one */
edit_package:
  call lineout tempfile,'<html><head><title>Edit an UpDater package </title></head>'
  call lineout tempfile,'<body 'back1'>'
  foo=sysfiletree(package_dir'\*.pkg','plist','FT')
  call lineout tempfile,'<h3>Edit an UpDater package </h3>'
  call lineout tempfile,'<ul>'
  call lineout tempfile,'<li> ServerName:<tt>'||servername||'</tt>'
  call lineout tempfile,'<li> Package directory: <tt>'||package_dir||'</tt><p>'
  call lineout tempfile,'</ul>'
  call lineout tempfile,'<form method="GET" action="'||thisaddon||'">'
  call lineout tempfile,'<input type="hidden" name="action" value="EDIT_PACKAGE_NEW">'
  call lineout tempfile,'Package name: <input type="text" name="name" size=22>'
  call lineout tempfile,'<input type="submit" value="Create the package">'
  call lineout tempfile,'</form><p>'


  call lineout tempfile,'<li> ... or choose a package to modify:'
  if plist.0<>0 then do
     call lineout tempfile,'<menu>'
     do pp=1 to plist.0
        n1=filespec('n',plist.pp)
        parse var n1 n1a '.' .
        call lineout tempfile,'<li><a href="'||thisaddon||'?action=EDIT_PACKAGE_MODIFY&package='||n1a||'">'||n1||'</a>'
     end
     call lineout tempfile,'</menu> '
   
   end

  call lineout tempfile,'</ul>'

  call lineout tempfile,'</BODY></HTML>'
  call lineout tempfile
  return 1


/*************************************/
edit_package2:
  call lineout tempfile,'<html><head><title>Create an UpDater package </title>'


  call lineout tempfile,'<script>'
  call lineout tempfile,' var  openargs=''scrollbars=yes,resizable=yes,screenX=50,screenY=50,innerWidth=400,toolbar=no,directories=no,dependent=yes'' ;' 
  call lineout tempfile,' var openargs2=''scrollbars=yes,resizable=yes,screenX=50,screenY=30,innerWidth=540,innerHeight=500,toolbar=no,directories=no,dependent=yes'' ;'
  call lineout tempfile,' var openargs3=''scrollbars=yes,resizable=yes,screenX=50,screenY=30,innerWidth=500,innerHeight=500,toolbar=no,directories=no,dependent=yes'' ;'
  call lineout tempfile,'</script>'


  call lineout tempfile,'</head>'
  call lineout tempfile,'<body 'back1'>'
  call lineout tempfile,'<h3>Create an UpDater package </h3>'

  sname=strip(strip(args.!NAME,,'\'))
  aname=package_dir'\'||sname
  parse var aname aname '.' .
  aname=aname||'.PKG'

  if stream(aname,'c','query exists')<>'' then do
       call lineout tempfile,'<table><tr><th bgcolor="lightyellow">Error</th>'
       call lineout tempfile,'<td>Package <tt>'||args.!name||'</tt> already exists. Try a different name'
       call lineout tempfile,'</td></table>'
       call lineout tempfile,'<a href="javascript:history.go(-1)">back</a>'
       signal isdone1
  end 

  call lineout tempfile,'<b>Creating package:</b> '
  call lineout tempfile,'<ul>'
  call lineout tempfile,'<li> ServerName:<tt>'||servername||'</tt>'
  call lineout tempfile,'<li> Package directory: <tt>'||package_dir||'</tt>'
  call lineout tempfile,'<li>              name: <tt>'||filespec('n',aname)||'</tt><p>'
  call lineout tempfile,'</ul>'

  call lineout tempfile,'<form method="POST" action="'||thisaddon||'">'
  call lineout tempfile,'<input type="hidden" name="action" value="EDIT_PACKAGE_NEW2">'
  call lineout tempfile,'<input type="hidden" name="name" value="'||sname||'">'

  call lineout tempfile,'<input type="submit" value="Create the package file"'
  call lineout tempfile,'<table><tr><th bgcolor="lightgrey">Comment </th>'
  call lineout tempfile,'<td> <input type="text" name="COMMENT" size=60></td>'

  call lineout tempfile,'<tr><th bgcolor="lightgrey">'
  call lineout tempfile,'  <a  target="cfg_doc" href="create_package_help" '
  goof1="'"||thisaddon||'?action=HELP_EDIT&var=TITLE'||"'"
  call lineout tempfile,' onClick="gpw=window.open('||goof1||',''cfg_doc'',openargs+'',innerHeight=478'');gpw.focus(); return false ; ">Title</a></td>'
  call lineout tempfile,'</th>'

  call lineout tempfile,'<td> <input type="text" name="TITLE" size=50></td>'

  call lineout tempfile,'<tr><th bgcolor="lightgrey">'
  call lineout tempfile,'  <a  target="cfg_doc" href="create_package_help" '
  goof1="'"||thisaddon||'?action=HELP_EDIT&var=Description'||"'"
  call lineout tempfile,' onClick="gpw=window.open('||goof1||',''cfg_doc'',openargs+'',innerHeight=478'');gpw.focus(); return false ; ">Description</a></td>'
  call lineout tempfile,'</th>'

  call lineout tempfile,'<td> <textarea name="description" rows=2 cols=50></textarea></td>'

  call lineout tempfile,'<tr><th bgcolor="lightgrey">'
  call lineout tempfile,'  <a  target="cfg_doc" href="create_package_help" '
  goof1="'"||thisaddon||'?action=HELP_EDIT&var=include'||"'"
  call lineout tempfile,' onClick="gpw=window.open('||goof1||',''cfg_doc'',openargs+'',innerHeight=478'');gpw.focus(); return false ; ">Include</a></td>'
  call lineout tempfile,'</th>'

/* INCLUDE */
  call lineout tempfile,'<td><table><th bgcolor="lightcyan">(relative) dir on<br> client</th>'
  call lineout tempfile,'<th>Dir\File specs on this server</th>'
  call lineout tempfile,'<th bgcolor="lightyellow">Include <br>subdirs</th>'
  do mmm=1 to 3
    call lineout tempfile,'<tr>'
    call lineout tempfile,'<td bgcolor="lightcyan"><input type="text" name="RELDIR_'||mmm||'" size=20></td>'
    call lineout tempfile,'<td><input type="text" name="ABSDIR_'||mmm||'" size=40></td>'
    call lineout tempfile,'<td bgcolor="lightyellow"><input type="checkbox" name="SUBDIR_'||mmm||'" value=1></td>'

  end
  call lineout tempfile,'</table></td>'

/* EXCLUDE */
  call lineout tempfile,'<tr><th bgcolor="lightgrey">'
  call lineout tempfile,'  <a  target="cfg_doc" href="create_package_help" '
  goof1="'"||thisaddon||'?action=HELP_EDIT&var=exclude'||"'"
  call lineout tempfile,' onClick="gpw=window.open('||goof1||',''cfg_doc'',openargs+'',innerHeight=478'');gpw.focus(); return false ; ">Exclude</a></td>'
  call lineout tempfile,'</th>'


  call lineout tempfile,'<td><table><th>Dir\File specs on this server</th>'
  call lineout tempfile,'<th bgcolor="lightyellow">Exclude subdirs</th>'

  do mmm=1 to 2
    call lineout tempfile,'<tr>'
    call lineout tempfile,'<td><input type="text" name="XABSDIR_'||mmm||'" size=40></td>'
    call lineout tempfile,'<td bgcolor="lightyellow"><input type="checkbox" name="XSUBDIR_'||mmm||'" value=1></td>'

  end
  call lineout tempfile,'</table></td>'

/* NO_Encode */
  call lineout tempfile,'<tr><th bgcolor="lightgrey">'
  call lineout tempfile,'  <a  target="cfg_doc" href="create_package_help" '
  goof1="'"||thisaddon||'?action=HELP_EDIT&var=No_Encode'||"'"
  call lineout tempfile,' onClick="gpw=window.open('||goof1||',''cfg_doc'',openargs+'',innerHeight=478'');gpw.focus(); return false ; ">No_Encode</a></td>'
  call lineout tempfile,'</th>'

  call lineout tempfile,'<td><table><th>Dir\File specs on this server</th>'
  do mmm=1 to 2
    call lineout tempfile,'<tr>'
    call lineout tempfile,'<td><input type="text" name="NOENCODE_'||mmm||'" size=40></td>'
  end
  call lineout tempfile,'</table></td>'


  call lineout tempfile,'<tr><th bgcolor="lightgrey">'
  call lineout tempfile,'  <a  target="cfg_doc" href="create_package_help" '
  goof1="'"||thisaddon||'?action=HELP_EDIT&var=Cache'||"'"
  call lineout tempfile,' onClick="gpw=window.open('||goof1||',''cfg_doc'',openargs+'',innerHeight=478'');gpw.focus(); return false ; ">Cache</a></td>'
  call lineout tempfile,'</th>'
  call lineout tempfile,'<td> <input type="text" name="CACHE"  size=5></td>'


  call lineout tempfile,'<tr><th bgcolor="lightgrey">'
  call lineout tempfile,'  <a  target="cfg_doc" href="create_package_help" '
  goof1="'"||thisaddon||'?action=HELP_EDIT&var=Run'||"'"
  call lineout tempfile,' onClick="gpw=window.open('||goof1||',''cfg_doc'',openargs+'',innerHeight=478'');gpw.focus(); return false ; ">Program to run</a></td>'
  call lineout tempfile,'</th>'
  call lineout tempfile,'<td> <input type="text" name="RUN"  size=45></td>'


  call lineout tempfile,'<tr><th bgcolor="lightgrey">'
  call lineout tempfile,'  <a  target="cfg_doc" href="create_package_help" '
  goof1="'"||thisaddon||'?action=HELP_EDIT&var=Privs'||"'"
  call lineout tempfile,' onClick="gpw=window.open('||goof1||',''cfg_doc'',openargs+'',innerHeight=478'');gpw.focus(); return false ; ">Required Privileges</a></td>'
  call lineout tempfile,'</th>'
  call lineout tempfile,'<td> <input type="text" name="PRIVS"  size=25></td>'


  call lineout tempfile,'<tr><th bgcolor="lightgrey">'
  call lineout tempfile,'  <a  target="cfg_doc" href="create_package_help" '
  goof1="'"||thisaddon||'?action=HELP_EDIT&var=Encrypt'||"'"
  call lineout tempfile,' onClick="gpw=window.open('||goof1||',''cfg_doc'',openargs+'',innerHeight=478'');gpw.focus(); return false ; ">Encrypt</a></td>'
  call lineout tempfile,'</th>'
  call lineout tempfile,'<td> <input type="text" name="ENCRYPT"  size=35></td>'

  call lineout tempfile,'</table>'

  call lineout tempfile,'<input type="submit" value="Create the package file"'

  call lineout tempfile,'</form>'
  call lineout tempfile,'<table><tr><th bgcolor="lightcyan">Note</th>'
  call lineout tempfile,'<td>To add extra <tt>Include, Exclude</tt> or <tt>No_Encode</tt>'
  call lineout tempfile,'entries, just edit this package (after saving it).</td></table>'

isdone1:
  call lineout tempfile,'</body></html>'
  call lineout tempfile
  return 1


/*************************************/
edit_package2a:
  call lineout tempfile,'<html><head><title>Save an UpDater package </title>'

  call lineout tempfile,'</head>'
  call lineout tempfile,'<body 'back1'>'
  call lineout tempfile,'<h3>Create an UpDater package (saving results) </h3>'

  sname=strip(strip(args.!NAME,,'\'))
  aname=package_dir'\'||sname
  parse var aname aname '.' .
  aname=aname||'.PKG'

  call lineout tempfile,'<ul>'
  call lineout tempfile,'<li> ServerName:<tt>'||servername||'</tt>'
  call lineout tempfile,'<li> Package directory: <tt>'||package_dir||'</tt>'
  call lineout tempfile,'<li>              name: <tt>'||filespec('n',aname)||'</tt><p>'

  if args.!TITLE='' | args.!RELDIR_1||args.!RELDIR_2||args.!RELDIR_3='' then do
          call lineout tempfile,'<li><b>ERROR</b> TITLE, or at least 1 INCLUDE, was not specified'
          call lineout tempfile,'</ul>'
          signal isdone1a
  end


/* backup current version of aname, if it exists */
  if stream(aname,'c','query exists')<>'' then do
       i1=lastpos('.',sname)
       if i1=0 then
          bname=sname||'.??'
       else
          bname=substr(sname,i1)||'??'
       bname=sre_tempfile(bname,'?',package_dir)
       foo=dosrename(aname,bname)
       if foo=1 then do
          call lineout tempfile,'<li><em>renaming old version to <tt>'bname||'</tt></em>'
       end
       else do
          call lineout tempfile,'<li><b>ERROR</b> Unable to rename old version of <tt>'aname'</tt> to '||bname
          call lineout tempfile,'</ul>'
          signal isdone1a
       end
  end

  call lineout tempfile,'</ul>'

/* now write lines to aname */

if args.!COMMENT<>'' then  call lineout aname,'; '||args.!comment

call lineout aname,'Title: '||args.!title
call lineout aname,' '

If args.!DESCRIPTION<>'' then do 
  call lineout aname,'Description:  '||args.!DESCRIPTION
  call lineout aname,' '
end

If args.!PRIVILEGES<>'' then do
  call lineout aname,'Privileges::  '||args.!PRIVILEGES
  call lineout aname,' '
end

call lineout aname,'Include: '
do jj=1 to 3
   a1='!RELDIR_'||jj
   a2='!ABSDIR_'||jj
   a3='!SUBDIR_'||jj
   IF STRIP(ARGS.A3)<>1 then
     CALL LINEOUT ANAME,'   '||strip(args.a1)||' '||args.a2
   else
     CALL LINEOUT ANAME,'   '||strip(args.a1)||' '||args.a2||' /S '
end
call lineout aname,' '


if args.!XABSDIR_1<>'' | args.!XABSDIR<>'' then do
   call lineout aname,'Exclude: '
  do jj=1 to 2
   a2='!XABSDIR_'||jj
   a3='!XSUBDIR_'||jj
   IF STRIP(ARGS.A3)<>1 then
     CALL LINEOUT ANAME,'   '||args.a2
   else
     CALL LINEOUT ANAME,'   '||args.a2||' /S '
 end
 call lineout aname,' '
end


if args.!NOENCODE_1<>'' | args.!NOENCODE_2<>'' then do
   call lineout aname,'No_encode: '
   do jj=1 to 2
      a2='!NOENCODE_'||jj
      call lineout aname,'   '||args.a2
   end
   call lineout aname,' '
end


If args.!CACHE<>'' then do
  if datatype(args.!cache)='NUM' then do
    call lineout aname,'Cache:  '||args.!Cache
    call lineout aname,' '
  end
end

If args.!RUN<>'' then do
  call lineout aname,'Run:  '||args.!RUN
  call lineout aname,' '
end

If args.!ENCRYPT<>'' then do
  call lineout aname,'Encrypt:  '||args.!ENCRYPT
  call lineout aname,' '
end

call lineout aname,' '
call lineout aname

iss=stream(aname,'c','query size')
call lineout tempfile,'<tt>'||iss||'</tt> bytes written to <u>'||sname||'</u> package file.'


isdone1a:
  call lineout tempfile,'</body></html>'
  call lineout tempfile
  return 1
  

/*************************************/
/* modify existing package */
edit_package3:
  call lineout tempfile,'<html><head><title>Modify an UpDater package </title>'

  call lineout tempfile,'<script>'
  call lineout tempfile,' var  openargs=''scrollbars=yes,resizable=yes,screenX=50,screenY=50,innerWidth=400,toolbar=no,directories=no,dependent=yes'' ;' 
  call lineout tempfile,' var openargs2=''scrollbars=yes,resizable=yes,screenX=50,screenY=30,innerWidth=540,innerHeight=500,toolbar=no,directories=no,dependent=yes'' ;'
  call lineout tempfile,' var openargs3=''scrollbars=yes,resizable=yes,screenX=50,screenY=30,innerWidth=500,innerHeight=500,toolbar=no,directories=no,dependent=yes'' ;'
  call lineout tempfile,'</script>'

  call lineout tempfile,'</head>'
  call lineout tempfile,'<body 'back1'>'
  call lineout tempfile,'<h3>Modify an UpDater package </h3>'

  sname=strip(strip(args.!PACKAGE,,'\'))
  aname=package_dir'\'||sname
  parse var aname aname '.' .
  aname=aname||'.PKG'

  isize=stream(aname,'c','query size')
  if isize=0 | isize='' then do
     call lineout tempfile,'<table><tr><th bgcolor="lightyellow">Error</th>'
     call lineout tempfile,'<td>Package <tt>'||args.!name||'</tt> does NOT exist: 'aname
     call lineout tempfile,'</td></table>'
     call lineout tempfile,'<a href="javascript:history.go(-1)">back</a>'
     signal isdone3
  end 

  stat=read_package(aname,isize)             /* read in package file to package_opts. */

/* package_opts set include
    !PRIVS !TITLE !DESC 
    .0 .n.!RDIR  .n.!CFILE .n.!SUBS      (includes)
    !EXCLUDE  !EXCLUDE.0  !EXCLUDE.n.1
    !NOENCODE.0  !NOENCODE.n  !NOENCODE.n.1
    !ENCRYPT  !RUN !CACHE
*/


  call lineout tempfile,'<b>Editing package:</b> '
  call lineout tempfile,'<ul>'
  call lineout tempfile,'<li> ServerName:<tt>'||servername||'</tt>'
  call lineout tempfile,'<li> Package directory: <tt>'||package_dir||'</tt>'
  call lineout tempfile,'<li>              name: <tt>'||filespec('n',aname)||'</tt><p>'
  call lineout tempfile,'</ul>'

  call lineout tempfile,'<form method="POST" action="'||thisaddon||'">'
  call lineout tempfile,'<input type="hidden" name="action" value="EDIT_PACKAGE_MODIFY2">'
  call lineout tempfile,'<input type="hidden" name="name" value="'||sname||'">'

  call lineout tempfile,'<input type="submit" value="Save the edits"'
  call lineout tempfile,'<table><tr><th bgcolor="lightgrey">Comment </th>'
  call lineout tempfile,'<td> <input type="text" name="NEWCOMMENT"  value="'||strip(package_opts.!COMMENT)||'" size=60></td>'

  call lineout tempfile,'<tr><th bgcolor="lightgrey">'
  call lineout tempfile,'  <a  target="cfg_doc" href="create_package_help" '
  goof1="'"||thisaddon||'?action=HELP_EDIT&var=TITLE'||"'"
  call lineout tempfile,' onClick="gpw=window.open('||goof1||',''cfg_doc'',openargs+'',innerHeight=478'');gpw.focus(); return false ; ">Title</a></td>'
  call lineout tempfile,'</th>'

  call lineout tempfile,'<td> <input type="text" name="TITLE" size=50  value="'||strip(package_opts.!TITLE)||'" ></td>'

  call lineout tempfile,'<tr><th bgcolor="lightgrey">'
  call lineout tempfile,'  <a  target="cfg_doc" href="create_package_help" '
  goof1="'"||thisaddon||'?action=HELP_EDIT&var=Description'||"'"
  call lineout tempfile,' onClick="gpw=window.open('||goof1||',''cfg_doc'',openargs+'',innerHeight=478'');gpw.focus(); return false ; ">Description</a></td>'
  call lineout tempfile,'</th>'

  call lineout tempfile,'<td> <textarea rows=2 cols=50>'||strip(package_opts.!DESC)||'</textarea></td>'


  call lineout tempfile,'<tr><th bgcolor="lightgrey">'
  call lineout tempfile,'  <a  target="cfg_doc" href="create_package_help" '
  goof1="'"||thisaddon||'?action=HELP_EDIT&var=Cache'||"'"
  call lineout tempfile,' onClick="gpw=window.open('||goof1||',''cfg_doc'',openargs+'',innerHeight=478'');gpw.focus(); return false ; ">Cache</a></td>'
  call lineout tempfile,'</th>'
  call lineout tempfile,'<td> <input type="text" name="CACHE"  value="'||package_opts.!CACHE||'"   size=5></td>'


  call lineout tempfile,'<tr><th bgcolor="lightgrey">'
  call lineout tempfile,'  <a  target="cfg_doc" href="create_package_help" '
  goof1="'"||thisaddon||'?action=HELP_EDIT&var=Privs'||"'"
  call lineout tempfile,' onClick="gpw=window.open('||goof1||',''cfg_doc'',openargs+'',innerHeight=478'');gpw.focus(); return false ; ">Required Privileges</a></td>'
  call lineout tempfile,'</th>'
  call lineout tempfile,'<td> <input type="text" name="PRIVS" value="'||strip(package_opts.!PRIVS)||'"  size=25></td>'

  call lineout tempfile,'<tr><th bgcolor="lightgrey">'
  call lineout tempfile,'  <a  target="cfg_doc" href="create_package_help" '
  goof1="'"||thisaddon||'?action=HELP_EDIT&var=Run'||"'"
  call lineout tempfile,' onClick="gpw=window.open('||goof1||',''cfg_doc'',openargs+'',innerHeight=478'');gpw.focus(); return false ; ">Program to run</a></td>'
  call lineout tempfile,'</th>'
  call lineout tempfile,'<td> <input type="text" name="RUN"  value="'||package_opts.!RUN||'"   size=45></td>'

  call lineout tempfile,'<tr><th bgcolor="lightgrey">'
  call lineout tempfile,'  <a  target="cfg_doc" href="create_package_help" '
  goof1="'"||thisaddon||'?action=HELP_EDIT&var=Encrypt'||"'"
  call lineout tempfile,' onClick="gpw=window.open('||goof1||',''cfg_doc'',openargs+'',innerHeight=478'');gpw.focus(); return false ; ">Encrypt</a></td>'
  call lineout tempfile,'</th>'
  call lineout tempfile,'<td> <input type="text" name="ENCRYPT" value="'||package_opts.!ENCRYPT||'"  size=35></td>'

  call lineout tempfile,'<tr><td colspan=2 bgcolor="lightblue" align="center"><br>Check to delete currently defined Include/Exclude/No_Encode entries </td>'

  call lineout tempfile,'<tr><th bgcolor="lightgrey">'
  call lineout tempfile,'<input type="checkbox" name="DEL_INCLUDE" value=1>'
  call lineout tempfile,'Include</th>'
  call lineout tempfile,'<td bgcolor="#ddddcc"><tt> '
  do i1=1 to package_opts.0
      call lineout tempfile,'<br>'||package_opts.i1.!rdir||' '||package_opts.i1.!CFILE
      if package_opts.i1.!SUBS=1 then call lineout tempfile,' /S '
  end
  call lineout tempfile,'</tt></td> '

  if package_opts.!Exclude.0>0 & package_opts.!Exclude.0<>'' then do
    call lineout tempfile,'<tr><th bgcolor="lightgrey">'
    call lineout tempfile,'<input type="checkbox" name="DEL_EXCLUDE" value=1>'
    call lineout tempfile,'Exclude</th>'
    call lineout tempfile,'<td bgcolor="#ddddcc"><tt> '
    do i1=1 to package_opts.!EXCLUDE.0
      call lineout tempfile,'<br>'||package_opts.!EXCLUDE.i1
      if package_opts.!EXCLUDE.i1.1=1 then call lineout tempfile,' /S '
    end
    call lineout tempfile,'</tt></td> '
  end


  if package_opts.!NOENCODE.0>0 & package_opts.!NOENCODE.0<>'' then do
    call lineout tempfile,'<tr><th bgcolor="lightgrey">'
    call lineout tempfile,'<input type="checkbox" name="DEL_NOENCODE" value=1>'
    call lineout tempfile,'No_Encode</th>'
    call lineout tempfile,'<td bgcolor="#ddddcc"><tt> '
    do i1=1 to package_opts.!NOENCODE.0
      call lineout tempfile,'<br>'||package_opts.!NOENCODE.i1
      if package_opts.!NOENCODE.i1.1=1 then call lineout tempfile,' /S '
    end
    call lineout tempfile,'</tt></td> '
  end


  call lineout tempfile,'<tr><td colspan=2 bgcolor="lightblue" align="center"><br>Additional Include/Exclude/No_Encode entries </td>'


  call lineout tempfile,'<tr><th bgcolor="lightgrey">'
  call lineout tempfile,'  <a  target="cfg_doc" href="create_package_help" '
  goof1="'"||thisaddon||'?action=HELP_EDIT&var=include'||"'"
  call lineout tempfile,' onClick="gpw=window.open('||goof1||',''cfg_doc'',openargs+'',innerHeight=478'');gpw.focus(); return false ; ">Include</a></td>'
  call lineout tempfile,'</th>'

/* INCLUDE */
  call lineout tempfile,'<td><table><th bgcolor="lightcyan">(relative) dir on<br> client</th>'
  call lineout tempfile,'<th>Dir\File specs on this server</th>'
  call lineout tempfile,'<th bgcolor="lightyellow">Include <br>subdirs</th>'
  do mmm=1 to 3
    call lineout tempfile,'<tr>'
    call lineout tempfile,'<td bgcolor="lightcyan"><input type="text" name="RELDIR_'||mmm||'" size=20></td>'
    call lineout tempfile,'<td><input type="text" name="ABSDIR_'||mmm||'" size=40></td>'
    call lineout tempfile,'<td bgcolor="lightyellow"><input type="checkbox" name="SUBDIR_'||mmm||'" value=1></td>'

  end
  call lineout tempfile,'</table></td>'

/* EXCLUDE */
  call lineout tempfile,'<tr><th bgcolor="lightgrey">'
  call lineout tempfile,'  <a  target="cfg_doc" href="create_package_help" '
  goof1="'"||thisaddon||'?action=HELP_EDIT&var=exclude'||"'"
  call lineout tempfile,' onClick="gpw=window.open('||goof1||',''cfg_doc'',openargs+'',innerHeight=478'');gpw.focus(); return false ; ">Exclude</a></td>'
  call lineout tempfile,'</th>'


  call lineout tempfile,'<td><table><th>Dir\File specs on this server</th>'
  call lineout tempfile,'<th bgcolor="lightyellow">Exclude subdirs</th>'

  do mmm=1 to 2
    call lineout tempfile,'<tr>'
    call lineout tempfile,'<td><input type="text" name="XABSDIR_'||mmm||'" size=40></td>'
    call lineout tempfile,'<td bgcolor="lightyellow"><input type="checkbox" name="XSUBDIR_'||mmm||'" value=1></td>'

  end
  call lineout tempfile,'</table></td>'

/* NO_Encode */
  call lineout tempfile,'<tr><th bgcolor="lightgrey">'
  call lineout tempfile,'  <a  target="cfg_doc" href="create_package_help" '
  goof1="'"||thisaddon||'?action=HELP_EDIT&var=No_Encode'||"'"
  call lineout tempfile,' onClick="gpw=window.open('||goof1||',''cfg_doc'',openargs+'',innerHeight=478'');gpw.focus(); return false ; ">No_Encode</a></td>'
  call lineout tempfile,'</th>'

  call lineout tempfile,'<td><table><th>Dir\File specs on this server</th>'
  do mmm=1 to 2
    call lineout tempfile,'<tr>'
    call lineout tempfile,'<td><input type="text" name="NOENCODE_'||mmm||'" size=40></td>'
  end
  call lineout tempfile,'</table></td>'


  call lineout tempfile,'</table>'

  call lineout tempfile,'<input type="submit" value="Save the edits"'

  call lineout tempfile,'</form>'
  call lineout tempfile,'<table><tr><th bgcolor="lightcyan">Note</th>'
  call lineout tempfile,'<td>To add extra <tt>Include, Exclude</tt> or <tt>No_Encode</tt>'
  call lineout tempfile,'entries, just edit this package (after saving it).</td></table>'

isdone3:
  call lineout tempfile,'</body></html>'
  call lineout tempfile
  return 1



/*************************************/
edit_package3a:
  call lineout tempfile,'<html><head><title>Save edits to an UpDater package </title>'

  call lineout tempfile,'</head>'
  call lineout tempfile,'<body 'back1'>'
  call lineout tempfile,'<h3>Edit an UpDater package (saving results) </h3>'

  sname=strip(strip(args.!NAME,,'\'))
  aname=package_dir'\'||sname
  parse var aname aname '.' .
  aname=aname||'.PKG'

  call lineout tempfile,'<ul>'
  call lineout tempfile,'<li> ServerName:<tt>'||servername||'</tt>'
  call lineout tempfile,'<li> Package directory: <tt>'||package_dir||'</tt>'
  call lineout tempfile,'<li>              name: <tt>'||filespec('n',aname)||'</tt><p>'


/* backup current version of aname, if it exists */
  if stream(aname,'c','query exists')<>'' then do
       i1=lastpos('.',sname)
       if i1=0 then
          bname=sname||'.??'
       else
          bname=substr(sname,i1)||'??'
       bname=sre_tempfile(bname,'?',package_dir)
       foo=dosrename(aname,bname)
       if foo=1 then do
          call lineout tempfile,'<li><em>renaming old version to <tt>'bname||'</tt></em>'
       end
       else do
          call lineout tempfile,'<li><b>ERROR</b> Unable to rename old version of <tt>'aname'</tt> to '||bname
          call lineout tempfile,'</ul>'
          signal isdone3a
       end
  end

  call lineout tempfile,'</ul>'

/* read in old version */
  ostuff=sre_read_file(bname,2,5,1)||'0d0a'x
  

/* copy lines from ostuff to aname, possibly replacing possibly removing */

if args.!COMMENT<>'' then  call lineout aname,'; '||args.!comment

afield=''; ministuff='' ; didlist=''
nblanks=0
do until ostuff=''
   parse var ostuff aline '0d0a'x ostuff
   if abbrev(strip(aline),';')=1 then do
      call lineout aname,aline
      nblanks=0
      iterate
   end
   if aline='' then do
      if afield='' then do
        nblanks=nblanks+1
        if nblanks<2 then call lineout aname,aline
        iterate
      end
      nblanks=0
      afield=strip(translate(afield))
      select
         when abbrev(afield,'INC') then afield='INCLUDE'
         when abbrev(afield,'EXC') then afield='EXCLUDE'
         when abbrev(afield,'NO_') then afield='NO_ENCODE'
         when abbrev(afield,'PRIV') then afield='PRIVS'
         when abbrev(afield,'ENC') then afield='ENCRYPT'
         when abbrev(afield,'DESC') then afield='DESCRIPTION'
         otherwise nop
      end
      vfield='!'||afield
/* INCLUDE, EXCLUDE, NO_ENCODE are special */
      select
         when afield='INCLUDE' then do
           if args.!DEL_INCLUDE<>1 then 
              call lineout aname,strip(afield)||': '||ministuff||'0d0a'x
           if args.vfield<>'' then
              call lineout aname,strip(afield)||':'||args.vfield||'0d0a'x
           didlist=didlist||' INCLUDE '
         end
         when afield='EXCLUDE' then do
           if args.!DEL_EXCLUDE<>1 then 
              call lineout aname,strip(afield)||': '||ministuff||'0d0a'x
           if args.vfield<>'' then
              call lineout aname,strip(afield)||':'||args.vfield||'0d0a'x
           didlist=didlist||' EXCLUDE '
         end

         when afield='NO_ENCODE' then do
           if args.!DEL_NOENCODE<>1 then 
              call lineout aname,strip(afield)||': '||ministuff||'0d0a'x
           if args.vfield<>'' then
              call lineout aname,strip(afield)||':'||args.vfield||'0d0a'x
           didlist=didlist||' NOENCODE '
         end

         otherwise do
           if args.vfield<>'' then
              call lineout aname,strip(afield)||':'||args.vfield||'0d0a'x
           else
              call lineout aname,strip(afield)||': '||ministuff||'0d0a'x
           didlist=didlist||' '||afield
         end
      end
      afield=''
      ministuff=''
      iterate
   end
   if afield<>'' then do
      if ministuff<>'' then
         ministuff=ministuff||'0d0a'x||aline
      else
         ministuff=aline
   end 
   else do
     if pos(':',aline)=0 then iterate   /* error, ignore */
     parse var aline afield ':' ministuff 
   end
end 

/* any other "ARGS" that haven't been done yet */

If args.!TITLE<>'' & wordpos('TITLE',didlist)=0 then do 
  call lineout aname,'Title:  '||args.!TIT:E
  call lineout aname,' '
end

If args.!DESCRIPTION<>'' & wordpos('DESCRIPTION',didlist)=0 then do 
  call lineout aname,'Description:  '||args.!DESCRIPTION
  call lineout aname,' '
end

If args.!PRIVILEGES<>''  & (wordpos('PRIVILEGES',didlist)+wordpos('PRIVS',didlist))=0 then do
  call lineout aname,'Privileges::  '||args.!PRIVILEGES
  call lineout aname,' '
end


a1='!RELDIR_1'
if args.a1<>'' then do
  call lineout aname,'Include: '
  do jj=1 to 3
   a1='!RELDIR_'||jj
   a2='!ABSDIR_'||jj
   a3='!SUBDIR_'||jj
   IF STRIP(ARGS.A3)<>1 then
     CALL LINEOUT ANAME,'   '||strip(args.a1)||' '||args.a2
   else
     CALL LINEOUT ANAME,'   '||strip(args.a1)||' '||args.a2||' /S '
  end
  call lineout aname,' '
end

if args.!XABSDIR_1<>'' | args.!XABSDIR<>'' &  wordpos('EXCLUDE',didlist)=0  then do
   call lineout aname,'Exclude: '
  do jj=1 to 2
   a2='!XABSDIR_'||jj
   a3='!XSUBDIR_'||jj
   IF STRIP(ARGS.A3)<>1 then
     CALL LINEOUT ANAME,'   '||args.a2
   else
     CALL LINEOUT ANAME,'   '||args.a2||' /S '
 end
 call lineout aname,' '
end


if args.!NOENCODE_1<>'' | args.!NOENCODE_2<>'' & wordpos('NOENCODE',didlist)=0  then do
   call lineout aname,'No_encode: '
   do jj=1 to 2
      a2='!NOENCODE_'||jj
      call lineout aname,'   '||args.a2
   end
   call lineout aname,' '
end


If args.!CACHE<>'' &  wordpos('CACHE',didlist)=0  then do
  if datatype(args.!cache)='NUM' then do
    call lineout aname,'Cache:  '||args.!Cache
    call lineout aname,' '
  end
end

If args.!RUN<>'' &  wordpos('RUN',didlist)=0  then do
  call lineout aname,'Run:  '||args.!RUN
  call lineout aname,' '
end

If args.!ENCRYPT<>'' &  wordpos('ENCRYPT',didlist)=0  then do
  call lineout aname,'Encrypt:  '||args.!ENCRYPT
  call lineout aname,' '
end

call lineout aname,' '
call lineout aname

iss=stream(aname,'c','query size')
call lineout tempfile,'<tt>'||iss||'</tt> bytes written to <u>'||sname||'</u> package file.'


isdone3a:
  call lineout tempfile,'</body></html>'
  call lineout tempfile
  return 1
  



/*************************************/
/* formatted display of  the various entries in this package */
disp_package:

  call lineout tempfile,'<html><head><title>Displaying an UpDater package </title></head>'
  call lineout tempfile,'<body 'back1'>'
  CALL LINEOUT TEMPFILE,'<h3>Contents of Updater Package: '||strip(translate(args.!package))'</h3>'
  call lineout tempfile,'<ul><li><b>Title</b>: 'package_opts.!title
  call lineout tempfile,'<li><b>Description</b>: 'package_opts.!desc
  if package_opts.!Privs<>'' then
     call lineout tempfile,'<li><b>Privileges</b>: 'package_opts.!privs
  call lineout tempfile,'</ul>'
  call lineout tempfile,'<h3>You can ... </h3>'

  torun=''
  if package_opts.!RUN<>'' & package_opts.!RUN<>'0'  then do
      parse var package_opts.!RUN torun . ; torun='torun='||strip(torun)||'&'
  end 

  call lineout tempfile,'<menu>'
  call lineout tempfile,'<li>'
  if (package_opts.!privs='' | pos('*',package_opts.!privs)>0)  & package_opts.!ENCRYPT='' then
    call lineout tempfile,'<a href="'||thisaddon||'?package='||args.!package||'&action=getl2&'||torun||'fst=1&client=1">'
  else
    call lineout tempfile,'<a href="'||thisaddon||'?package='||args.!package||'&action=getl2&'||torun||'fst=1&client=1&GETUSER=1">'
  call lineout tempfile,'UpDate these files</a> to your computer '
  call lineout tempfile,'(using the <a target="info" href="'||thisaddon||'?action=UPCLIENTDOC">UpClient helper app</a>) '
  call lineout tempfile,'<li> or ... '
  if (package_opts.!privs='' | pos('*',package_opts.!privs)>0) & package_opts.!ENCRYPT='' then
    call lineout tempfile,'<a href="'||thisaddon||'?package='||args.!package||'&action=getl3&'||torun||'fst=1&client=1">'
  else
    call lineout tempfile,'<a href="'||thisaddon||'?package='||args.!package||'&action=getl3&'||torun||'fst=1&client=1&GETUSER=1">'
  call lineout tempfile,'selectively UpDate these files</a> to your computer '

  call lineout tempfile,'<li>List all the '
  call lineout tempfile,'<a href="'||thisaddon||'?package='||args.!package||'&action=getl&ashtml=1">files',
                       ' in 'args.!package '</a>?'
  call lineout tempfile,'<li><a href="'||thisaddon||'?">Return to UpDater main menu</a> (at 'servername')'

  call lineout tempfile,'</menu>'
  call lineout tempfile,'<h3>Description of the 'args.!package' package</h3>'


  call lineout tempfile,'<ul> '
     if package_opts.!PRIVS<>'' then call lineout tempfile,'<li>User-specific privileges required '
     if package_opts.!ENCRYPT<>'' then call lineout tempfile,'<li>Encryption methods available: <tt> 'package_opts.!ENCRYPT '</tt>'
     if package_opts.!RUN<>'' & package_opts.!RUN<>'0' then call lineout tempfile,'<li>Program to run after UpDating:<tt> 'package_opts.!RUN '</tt>'
  call lineout tempfile,'</ul>'

  call lineout tempfile,'<table border=1><tr><th>Remote Dir </th><th>Central-server Files </th><th>Subdirectories? </th>'

  do jj= 1 to package_opts.0
     sok='' 
     if package_opts.jj.!subs=1 then sok='yes'
     a1=package_opts.jj.!rdir
     a2=package_opts.jj.!cfile
     call lineout tempfile,'<tr><td> 'a1'</td><td> 'a2'</td><td> 'sok'</td></tr>'
  end
  call lineout tempfile,'</table>'
  if package_opts.!exclude.0>0  then do
    call lineout tempfile,'<hr width="30%">'
    call lineout tempfile,'<b>File exclusion rules</b><br>'
    call lineout tempfile,'<table border=1><tr><th>File Rule</th><th>Subdirectories? </th>'
    do jj= 1 to package_opts.!exclude.0
       sok='' 
       if package_opts.!exclude.jj.1=1 then sok='yes'
       a2=package_opts.!exclude.jj
       call lineout tempfile,'<tr><td> 'a2'</td><td> 'sok'</td></tr>'
    end 
  end
  call lineout tempfile,'</table>'
  call lineout tempfile,'</body></html>'
  call lineout tempfile
  return 1

  end

/***********/
/* describe the upclient helper app */
tell_upclient:
  call lineout tempfile,'<html><head><title>The UpClient helper app </title></head>'
  call lineout tempfile,'<body 'back1'>'
  call lineout tempfile,'<a name="upclient"><h3>The UpClient helper app </h3></a>'
  call lineout tempfile,'The <a href="'||thisaddon||'?send=UPCLIENT.CMD">UpClient</a> helper application is a REXX program (UPCLIENT.CMD)'
  call lineout tempfile,'that will update a package of files to your computer. '
  call lineout tempfile,'<b>UPCLIENT.CMD is a helper application for OS/2 Netscape, or IBM OS/2 Web Browser, or the latest OS/2 Mozilla.</b>'
  call lineout tempfile,'<p>Assuming this describes you, after '
  call lineout tempfile,'<a href="'||thisaddon||'?send=UPCLIENT.CMD">downloading UpClient</a>'
  call lineout tempfile,'you should  (the following assumes you are using Netscape)...<ol>'
  call lineout tempfile,'<li>Copy UPCLIENT.CMD  to a directory in your PATH (i.e.; to x:\OS2\APPS, where x: is your boot drive)'
  call lineout tempfile,'<li>Tell NetScape (or IBM Web Browser) to use UPCLIENT.CMD with all <tt>application/x-updater</tt> responses'
  call lineout tempfile,'<br>To do this, you should: '
  call lineout tempfile,'<menu><li>Open up the Preferences menu (or the <tt> Edit-Preferences</tt> menu)'
  call lineout tempfile,'<li><em>Netscape:</em> Select <tt>Applications</tt>'
  call lineout tempfile,'<br><em>IBM Web Browser:</em> Select <tt>Helper Applications</tt>'

  call lineout tempfile,'<li>Select the <tt>New Type</tt> option'
  call lineout tempfile,'<li>Fill in the fields of the <tt>New Type</tt> screen with: '
  call lineout tempfile,'<ul><li><u>Description of type</u>: UpDater Client '
  call lineout tempfile,'<li><u>File Extension</u>: UPD '
  call lineout tempfile,'<li><u>MIME Type</u>: application/x-updater'
  call lineout tempfile,'<li>'
  call lineout tempfile,'<table ><tr><td><u>Application to use</u>:</td>'
  call lineout tempfile,'<td><Menu><li><Em>Netscape:</em> <tt> cmd.exe /c "x:\os2\apps\upclient.cmd"</tt>'
  call lineout tempfile,'<br> <em>be sure to include the " characters </em>'
  call lineout tempfile,'<li><Em>IBM Web Browser:</em> <tt> x:\os2\apps\upclient.cmd</tt></menu>'
  call lineout tempfile,'</td></table>'

  call lineout tempfile,'</ul>'
  call lineout tempfile,'</menu>'
  call lineout tempfile,'</ol>'
  call lineout tempfile,'Once UpClient is installed, when you UpDate a package from a'
  call lineout tempfile,'<Em>UpDater central server</em>, Netscape will run UpClient in '
  call lineout tempfile,'seperate window.  After asking a few questions, UpClient will'
  call lineout tempfile,'GET files and write them to your hard drive.'
  call lineout tempfile,'<p>Alternatively, <a href="'||thisaddon||'?send=updateme.cmd">UpDateMe.cmd</a>, '
  call lineout tempfile,'a non-gui stand-alone program, can be used as your client-side interface'
  call lineout tempfile,'to a central server.'

  call lineout tempfile,'<hr width="30%">'
  call lineout tempfile,'Note: UpClient works best when you have the following libraries installed:'
  call lineout tempfile,'<ol>'
  call lineout tempfile,'<li> <a href="'||thisaddon||'?send=rxRsync.dll">The rxRsync library</a>'
  call lineout tempfile,'<li> <a href="'||thisaddon||'?send=rxgzlib.dll">The rxgzlib library</a>'
  call lineout tempfile,'<p><em>(install these .dll files in your <TT>LIBPATH</tt>. For example, in x:\OS2\DLL)</em>'

  call lineout tempfile,'</ol>'

  call lineout tempfile,'</body></html>'

  call lineout tempfile
  return 1

/*************************************/
/* check to see if client (remote server) has rights to this package */

check_privs:procedure expose tempfile args. servername  thisaddon id_info rcode aiter
parse upper arg privset,daprivs

IF WORDPOS('SUPERUSER',PRIVSET)>0 then RETURN 'OK' 
do mm=1 to words(daprivs)
   ap=strip(word(daprivs,mm))
   if wordpos(ap,privset)>0 then return 'OK'
end
if aiter='' then do
   if daprivs='' | wordpos('*',daprivs)>0 then
      rcode='PRIVS_LOOK  , To use UpDater''s encryption facility, you need to provide a username and password known to this system'
   else
      rcode='PRIVS_CHECK  , To use obtain  this  package, UpDater requires user user specific privileges'
   return 1
end 

 if daprivs='' | wordpos('*',daprivs)>0 then return 'OK'

 message='<!doctype html public "-//IETF//DTD HTML 3.0//EN">'
 message=message||'<html><head><title>UpDater. Please enter a valid username</title></head> '
 message=message||' <strong>'
 message=message||' To acquire 'args.!package', a valid  username is required</strong>'
 message=message||' </body> </html> '
rcode=sre_auth_response('UPDATER',servername,1,message,id_info)
return 1

/*************************************/
/* read package file into package_opts. stem */
read_package:procedure expose package_opts. crlf candorsync candogzip  thisaddon
parse arg afile,isize

/* package_opts set include
    !PRIVS !TITLE !DESC 
    .0 .n.!RDIR  .n.!CFILE .n.!SUBS      (includes)
    !EXCLUDE  !EXCLUDE.0  !EXCLUDE.n.1
    !NOENCODE.0  !NOENCODE.n  !NOENCODE.n.1
    !ENCRYPT  !RUN !CACHE
    !COMMENT

*/


afilea=charin(afile,1,isize)||'0d0a'x||' '
foo=stream(afile,'c','close')

doing=''
package_opts.=''
package_opts.!exclude.0=0
package_opts.!noencode.0=0
package_opts.!cache=0

parse var afile afil0 '.' .
afilc=afil0'.PKC'

if stream(afilc,'c','query size')=0 then foo=sysfiledelete(afilc)  /* 0 length files are trouble */

package_opts.!PKG_DATE=make_timestamp(afile)
package_opts.!cachefile=afilc
package_opts.!cachestamp=make_timestamp(afilc)  /* timestamp of "cache" entry */

package_opts.0=0

do until afilea=''
  parse var afilea aline (crlf) afilea
  aline=strip(aline)

  select 
    when abbrev(aline,';')=1 then do 
        package_opts.!COMMENT=package_opts.!COMMENT||' '||strip(strip(aline),'l',';')
    end 
    when aline='' & doing='' then iterate   /* skip multiple blank lines */
    when aline='' then do           /* end of a entry */ 
       doing=''                   /* get ready for next entry */
       iterate
    end
    when doing='' then do           /* find name: */
      if pos(':',aline)=0 then iterate  /* error, ignore */
      parse var aline aname ':' aline
      doing=strip(translate(aname))
      
    end
    OTHERWISE NOP
   END
   if aline='' then iterate
   SELECT
    when abbrev(doing,'PRIV')=1 then do   /* not a FILES entry */
        package_opts.!PRIVS=package_opts.!PRIVS' 'aline
        iterate
    end
    when abbrev(doing,'DESC')=1 then do   /* not a FILES entry */
        package_opts.!DESC=package_opts.!DESC' 'aline
        iterate
    end
    when abbrev(doing,'TITLE')=1 then do   /* not a FILES entry */
        package_opts.!TITLE=package_opts.!TITLE' 'aline
        iterate
    end

    when abbrev(doing,'FILE')+abbrev(doing,'INC')>0 then do   /* an include entry */
       parse var aline remote_dir central_file alsosubs .
       inn=package_opts.0+1
       package_opts.inn.!rdir='/'||strip(translate(strip(translate(remote_dir)),'/','\'),,'/')

       ifile=translate(strip(translate(central_file)),'\','/')

       foo=translate(ifile,' ','\')
       aa=word(foo,words(foo))
       if (pos('.',aa)+pos('*',aa)+pos('?',aa))=0 then ifile=ifile'\*.*'

       package_opts.inn.!cfile=ifile
       if translate(strip(alsosubs))='/S' then package_opts.inn.!subs=1
       package_opts.0=inn
    end

    when abbrev(doing,'EXC')=1 then do   /* a EXCLUDE entry */
       parse var aline  exc_file alsosubs .
       inn=package_opts.!exclude.0+1
       efile=translate(strip(translate(exc_file)),'\','/')

       foo=translate(efile,' ','\')
       aa=word(foo,words(foo))
       if (pos('.',aa)+pos('*',aa)+pos('?',aa))=0 then efile=efile'\*.*'

       package_opts.!exclude.inn=efile
       if translate(strip(alsosubs))='/S' then package_opts.!exclude.inn.1=1
       package_opts.!exclude.0=inn

    end

    when abbrev(doing,'NO_E')=1 then do   /* a no_encode entry */
       parse var aline  noe_file alsosubs  .
       inn=package_opts.!noencode.0+1
       package_opts.!noencode.inn=translate(strip(translate(noe_file)),'\','/')
       if translate(strip(alsosubs))='/S' then package_opts.!noencode.inn.1=1
       package_opts.!noencode.0=inn
    end

    when abbrev(doing,'ENCRYPT')=1 then do
        package_opts.!encrypt=translate(strip(space(aline)))
    end 

    when abbrev(doing,'RUN')=1 then do
        package_opts.!run=strip(aline)
    end 


    when abbrev(doing,'CACHE')=1 then do   /* a no_encode entry */
       parse var aline  adays . ; adays=strip(adays)
       if datatype(adays)='NUM' then  package_opts.!cache=adays
    end

    otherwise do
      AOPT='!'||DOING
      package_opts.aopt=package_opts.aopt||' '||aline
      iterate
    end 

  end         /*select */
end     
return 1


/****/
/* make a nnn.xxx timestamp */
make_timestamp:procedure
parse arg afile
foo=sysfiletree(afile,'ttt','FT')
if ttt.0=0 then return 0                /* no time stamp */
parse var ttt.1 yr '/' mon '/' day '/' hr '/' mnutes .
if yr<70 then 
  yr=2000+yr
else
  yr=1900+yr
tt2=dateconv(yr||mon||day,'S','B')
numeric digits 10
tt3=((hr*60)+mnutes)/1440
tstamp=tt2+tt3
return tstamp 

/***********/
/* adler-32 crc of file aname */
fileadler:procedure expose candorsync
parse arg aname

if candorsync=0 then return ''

foo=stream(aname,'c','open read')
if abbrev(strip(translate(foo)),'READY')=0 then return ''

sz=stream(aname,'c','query size')
if sz='' | sz=0 then return ' '
aa=charin(aname,1,min(100000,sz))
adler=rx_adler32(aa)
igot=100000
do forever
   if igot>=sz then leave
   iget=min(100000,sz-igot)
   aa=charin(aname,,iget)
    adler=rx_adler32(aa,adler)
    igot=igot+100000
end
foo=stream(aname,'c','close')

return adler



/*** jump here on error */
badguys:
say "updater error "Rc " at " sigl
if sre_completed()=0 then do

    foo=sre_command('string error 'rc ' at ' sigl,,,,'x-updater: error')
end

return ' '


/********************/
/* check for gzip and rsync dlls -- note that we assume rxsock, rexxutil,
   and rexxlib are available */

check_dlls:

if rxfuncquery('rx_md4')=1  then do
  foo=RXFuncAdd( 'RXRsyncLoad', 'RXRSYNC', 'RxRsyncLoad')
  if foo=0 then call RxRsyncLoad
end
if rxfuncquery('rx_md4')=1  then do
   candorsync=0
   call pmprintf("UpDater warning: unable to find rxrsync")
end


return 0

/********************/
/* popup help for edit_package options */
help_edit:


call lineout tempfile,'<html><head><title>Edit-Package help </title></head>'
call lineout tempfile,'<body 'back1'>'
call lineout tempfile,'<font size=-1><a href="javascript:close()">Close</a></font>'
call lineout tempfile

call lineout tempfile,'<dl>'

avar=strip(translate(args.!VAR))

select


   when avar='CACHE' then do
    call lineout tempfile,'<dt><b>Cache</b> <em>optional</em>'
    call lineout tempfile,'<dd>Days to retain "list of files"<br>' 
    call lineout tempfile,'    Number of days to retain the "list of files" (in this package).'
    call lineout tempfile,'    To reduce the amount of time spent generating "list of files" from the definitions found in a package, you can specify the number'
    call lineout tempfile,'    of days to use a "cached" version of this list. For rapidly changing packages this is not a good idea, but for rarely changing packages'
    call lineout tempfile,'    that contain lots of files the use of cache can reduce processing time (hence speed up throughput). '
   end



   when avar='DESCRIPTION' then do
    call lineout tempfile,'<dt><b>Descprition</b> <em>optional</em>'
    call lineout tempfile,'<dd> A long, typically multi-line, description of this package. '
   end

   when avar='ENCRYPT' then do
    call lineout tempfile,'<dt><b>Encrypt</b> <em>optional</em>'
    call lineout tempfile,'<dd> a space delimited list of supported encryption methods' 
    call lineout tempfile,' These are encryption methods that the central server may apply to the files in this package. ' 
    call lineout tempfile,' <br>Note: this server supports the following encryption methods: <tt>'
    goo3=cvtails('ENCRYPT.','goo4')
    goo5='#SIMPLE '
    do gg=1 to goo4.0
       goo5=goo5||' '||substr(translate(strip(goo4.gg)),2)
    end 
    if goo5='' then
      call lineout tempfile,goo5||'No encryption methods are currently supported.</tt>'
    else
      call lineout tempfile,goo5||'</tt>'
   end


   when avar='EXCLUDE' then do
    call lineout tempfile,'<dt><b>Exclude  </b> <em>optional</em>'
    call lineout tempfile,'<dd>a list of files to exclude.<br> '
    call lineout tempfile,'    A list of files to exclude, with * wildcards; you can also instruct UpDater to exclude subdirectories.'
    call lineout tempfile,'   Files that match an <tt>include</tt>'
    call lineout tempfile,'    entry are checked against <tt>exclude</tt> entries -- if a matching exclude is found, the file is <b>not</b> included.  ' 
   end



   when avar='INCLUDE' then do
    call lineout tempfile,'<dt><b>Include  </b> <em>required</em>'
    call lineout tempfile,'<dd>  A list of files in this package.<br>'
    call lineout tempfile,'    This is a list of files that comprise this package. Each item is specified by two components: a'
    call lineout tempfile,'    relative destination directory, and a fully qualified file name.'
    call lineout tempfile,'    <br>The fully qualified file name can include wildcards (* characters). '
    call lineout tempfile,'        In addition, you can also include  files in subdirectories.'
    call lineout tempfile,'    <br>Basically, UpDater will copy the files on a central server (as identified by these filenames) to a relative (to a base-directory you'
    call lineout tempfile,'    specify) directory on a remote server. ' 
   end

   when avar='NO_ENCODE' then do
    call lineout tempfile,'<dt><b>NO_ENCODE  </b> <em>optional</em>'
    call lineout tempfile,'<dd>a list of files to NOT content-encode<br>' 
    call lineout tempfile,'    A list of files, using the same syntax as exclude. Files in no_encode will not be GZIP''ed or RSYNC''ed. This is especially useful'
    call lineout tempfile,'    when you have .ZIP files, or other compressed files -- since further compression rarely helps, and differences are often difficult to'
    call lineout tempfile,'    detect.  ' 
   end



   when avar='PRIVILEGES' then do
    call lineout tempfile,'<dt><b>Privileges</b> <em>optional</em>'
    call lineout tempfile,'<dd> SREhttp/2 privileges required to obtain this package. <br>'
    call lineout tempfile,'    These are user-specific privileges that are set by SREhttp/2. Note that these are in addition to any privileges that SREHttp/2 may'
    call lineout tempfile,'    require on a selector specific basis. In other words, you might have a general set of privileges that permit access to the UpDater'
    call lineout tempfile,'    addon, and a variety of specific privileges required for each of several packages. '
    call lineout tempfile,'     <dl><dt>Note:<dd> '
    call lineout tempfile,'         If you do not want to limite access, you should <b>not</b> specify <tt>privileges</tt></dl>.'
   end

   when avar='RUN' then do
    call lineout tempfile,'<dt><b>Privs</b> '
    call lineout tempfile,'<dd> Execute an installation program on a remote client. <br>'
    call lineout tempfile,'    You can specify a program to run after the helper application has completed. This will only be run if the client permits it --'
    call lineout tempfile,'    which requires BOTH checking a box in a form, and answering Yes to a "do you want to run" question asked by the helper'
    call lineout tempfile,'    application. '
    call lineout tempfile,'    <dL><dt>The syntax of run: is ... '
    call lineout tempfile,'     <dd><tt>[relative_path]program [PM] [arg=a sting]</tt> </dl>'
    call lineout tempfile,'<br><em>Note:</em> to <b>not </b>  run a program, enter <tt>0</tt>' 
  end


   when avar='PRIVS' then do
    call lineout tempfile,'<dt><b>Run</b> '
    call lineout tempfile,'<dd> SREhttp/2 privileges required to obtain this package.<br>'
    call lineout tempfile,'    These are user-specific privileges that are set by SREhttp/2.'
    call lineout tempfile,' Note that these are in addition to any privileges that SREHttp/2 '
    call lineout tempfile,'may require simply to access the UPDATER addon (on this <tt>central server</tt>)'
    call lineout tempfile,' <br>In other words, you might have a general set of privileges that '
    call lineout tempfile,' permit access to the UpDater addon, and a variety of specific privileges '
    call lineout tempfile,'required for each of several packages.'
    call lineout tempfile,'<br><b>Note:</b> If you do not want to limit access, you should <b>not</b> specify any <tt>privileges</tt>.'
  end

   when avar='TITLE' then do
    call lineout tempfile,'<dt><b>Title</b> <em>required</em>'
    call lineout tempfile,'<dd> The title of the package.<br>'
    call lineout tempfile,'    This should be a fairly short (50 character) name for this package. It will be used by "remote servers" when communicating with'
    call lineout tempfile,'    clients. '
   end


otherwise
end  

call lineout tempfile,'</dl>'
call lineout tempfile,'</body></html>'
call lineout tempfile
return  1



/* fix up size, given format */
fixup_size:procedure
parse arg asize,isabbrev
if isabbrev=1 then do
   if asize>=1000000 then
        return format(asize/1000000,,1)||'M'
   if asize>=1000 then 
      return format(asize/1000,,0)||'K'
end
/* convert to xxx,yyy,zzz */
 il=length(asize)
 if il>3 then do
           oop=""
           do mm=il to 3 by -3
               tt=substr(asize,mm-2,3)
               if mm=il then
                  oop=tt
               else
                 oop=tt||','||oop
           end /* do */
           if mm<>0 then oop=substr(asize,1,mm)||','||oop
           asize=oop
 end
 return asize    /* not abbrev, or < 1000 */



/******************/
/* access allowed? */
sec_check:
ss=sre_value('SECURITY_LEVEL',,'SRESYS')
if ss=0 then do
  rcode=sre_command('String Sorry. Remote access to configuration utilities not permitted.',,idinfo)
  return 0
end
if ss=1 then do
    myaddr=sre_extract("clientaddr",idinfo)
    moreaddr=sre_value('SERVERADDR_INHOUSE',,'SRESYS')
    who=sre_extract("serveraddr",idinfo)||' '||moreaddr
    if wordpos(myaddr,who)=0 then do
      rcode=sre_command('String External access to configuration utilities is not permitted' ,,idinfo)
      return 0
   end
end
/* else, normal superuser checking */
crlf='0d0a'x
parse upper var uri action '?' .
if wordpos('SUPERUSER',privset)=0 then do
  if aiter='' then do
    rcode='PRIVS_CHECK SUPERUSER , 'action' addon requires superuser privileges '
    return 0
  end
  rcode='AUTH 'action'  addon requires that you have superuser privileges '
  return 0
end 
return 1



